#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
File handling script for formatting output files, getting file lists, and
reading and writing to database containing:
    get_file_list(out_dir,begin):       input GENE output directory and base filepath 
                                           (nrg, energy, etc), return full list of files 
                                           in directory
    get_suffixes(out_dir):            input GENE output directory, return list of run 
                                           suffixes in the directory
    gridfs_put(filepath):               input filepath, upload  file to database, and 
                                           return object_id of uploaded file
    gridfs_read(db_file):               input database filename, return contents of file
    upload_to_mongo   
    isLinear
@author: Austin Blackmon, Dongyang Kuang, Venkitesh Ayyar
"""

import sys
import numpy as np
from bson.objectid import ObjectId
import os
from pathlib import Path
import gridfs
import json
import yaml
from time import strftime
import pickle
from bson.binary import Binary

from .pyro_gk import create_gk_dict_with_pyro
from .ParIO import Parameters
from .diag_plot import diag_plot
from .mgk_post_processing import get_parsed_params, get_suffixes, get_diag_from_run

#=======================================================

class Global_vars():
    '''
    Object to store global variables
    '''
    def __init__(self, sim_type):

        self.set_vars(sim_type)
        #User specified files#
        self.Docs_ex = [] 
        self.update_docs_keys()
        self.troubled_runs = [] # a global list to collection runs where exception happens

    def set_vars(self, sim_type):
        if sim_type=="GENE":

            self.required_files = ['field', 'nrg', 'omega','parameters']
            self.Docs         = ['autopar', 'nrg', 'omega','parameters']
            ## Geometry files added later in processing

            #Large files#
            self.Docs_L = ['field', 'mom', 'vsp']

        elif sim_type=='CGYRO':

            self.required_files =  ['input.cgyro', "out.cgyro.time","out.cgyro.grids","out.cgyro.equilibrium", "bin.cgyro.geo"]
            
            self.Docs = ['input.cgyro', 'input.cgyro.gen', 'input.gacode', 'out.cgyro.info']    
            
            #Large files#
            self.Docs_L = []

        elif sim_type=='TGLF':
            self.required_files = ['input.tglf', 'out.tglf.run']    

            self.Docs = ['input.tglf', 'input.tglf.gen', 'out.tglf.run']    

            #Large files#
            self.Docs_L = []

        elif sim_type=='GX':

            self.required_files = ['gx.in','gx.out.nc']    

            self.Docs = ['gx.in']    

            #Large files#
            self.Docs_L = []

        elif sim_type=='GS2':
            self.required_files = ['gs2.in','gs2.out.nc']    

            self.Docs = ['gs2.in']    

            #Large files#
            self.Docs_L = []
        
        else : 
            print("Invalid simulation type",sim_type)
            raise SystemError 
        
        ### Keys used for filenames 
        self.Keys = [fname.replace('.','_') for fname in self.Docs]

    def update_docs_keys(self):

        self.all_file_docs = self.Docs + self.Docs_ex
        ## Drop any duplicates 
        self.all_file_docs = list(set(self.all_file_docs))

        self.all_file_keys =  [fname.replace('.','_') for fname in self.all_file_docs]
    
    def reset_docs_keys(self,sim_type):
        ## Reset values 
        self.set_vars(sim_type)
        self.update_docs_keys() ## Update extra files
        print("File names and their key names are reset to default!")

def f_load_config(config_file):
    '''
    Load config file instead of user prompts
    '''
    
    with open(config_file) as f:
        config_dict=yaml.load(f, Loader=yaml.SafeLoader)
    
    return config_dict 

def f_check_required_files(global_vars, fldr, suffix, sim_type):

    files_exist=True 

    for fname in global_vars.required_files:
        file = os.path.join(fldr,fname+suffix) if sim_type=='GENE' else os.path.join(fldr,suffix,fname)

        if not os.path.isfile(file) :
            print('Necessary file %s does not exist. Skipping this suffix %s'%(file,suffix))
            files_exist = False
            break
        elif (os.path.getsize(file) == 0):
            print('Necessary file %s is empty. Skipping this suffix %s'%(file,suffix))
            files_exist = False
            break
    
    return files_exist

def f_user_input_metadata(database):
    '''
    Create a dictonary of user inputs for metadata
    Used as keyword arguments to construct metadata dictionary
    '''

    user_ip = {} 
    print("Filling metadata.")
    skip_metadata = input("To skip entering Metadata, please enter 0. press any other key to continue\n")
    if skip_metadata=='0': 
        return user_ip

    print("Please provide input for metadata. Press Enter to skip that entry.\n")
    confidence = input('What is your confidence (1-10) for the run? Press ENTER to use default value -1.0\n')
    if len(confidence):
        confidence = float(confidence)
    else:
        confidence = -1.0
        print("Using default confidence -1.\n")

    user_ip['confidence']= confidence 

    keywords = input('Any keywords to categorize this run. Press Enter to skip\n')
    user_ip['keywords'] = keywords
    
    comments = input('Any comments for data in this folder?Press Enter to skip.\n')
    user_ip['comments'] = comments

    linked_id_strg = input('Do you want to link this run to another existing run in MGKDB. Press Enter to skip\n')
    user_ip['linked_ID'] = f_get_linked_oid(database, linked_id_strg) if linked_id_strg else None

    archive = input('Is there a location where the data is archived? Press Enter to skip.\n')
    user_ip['archive_loc'] = archive

    restart = input('Is this run a restart starting from a different run? For yes -> Y . no -> any other key \n')

    user_ip['restart'] = (restart=='Y')

    if restart=='Y':
        user_ip['restart_timestep'] = int(input('What was the timestep of the previous run used to start this run?\n'))
        initial_run_info = input('Has the initial run been uploaded to this database. For yes -> Y .\n')
        run_oid = ObjectId(input('Please enter the ObjectID for that run\n')) if initial_run_info == 'Y' else None
        if f_check_id_exists(database, run_oid):
            user_ip['initial_run_oid'] =  run_oid 
        else:
            print(f"Entered object id {run_oid} doesn't exist\n")
    
    expt = input('Name of actual or hypothetical experiment? Eg: diiid, iter, sparc, etc. Press Enter to skip.\n')
    user_ip['expt'] = expt

    scenario_id = input('Scenario ID : shot ID or time or runID? Eg: 129913.1500ms . Press Enter to skip.\n')
    user_ip['scenario_runid'] = scenario_id

    git_hash = input('Do you have git-hash to store?Press Enter to skip.\n')
    user_ip['git_hash'] = git_hash

    platform = input('Platform on which this was run? Eg: perlmutter, summit, engaging, pc . Press Enter to skip.\n')
    user_ip['platform'] = platform

    exec_date = input('Execution date?Press Enter to skip.\n')
    user_ip['ex_date'] = exec_date

    workflow = input('Workflow type? Eg: portals, smarts, standalone, etc. Press Enter to skip.\n')
    user_ip['workflow_type'] = workflow

    print("Publication information should be uploaded with a separate script")

    return user_ip

def f_set_metadata(user=None,out_dir=None,suffix=None,keywords=None,confidence=-1,comments='Uploaded with default settings.',time_upload=None,\
                   last_update=None, linked_ID=None, expt=None, scenario_runid=None, linear=None, quasiLinear=None, has_1dflux = None, sim_type=None,\
                   git_hash=None, platform=None, ex_date=None, workflow_type=None, archive_loc=None, restart=False, restart_timestep=0, initial_run_oid=None):

    metadata={
        'DBtag': { 
            'user': user,
            'run_collection_name': out_dir,
            'run_suffix': suffix,
            'keywords':keywords,
            'confidence': confidence,
            'comments': comments,
            'time_uploaded': time_upload,
            'last_updated': last_update,
            'linkedObjectID': linked_ID, 
            'archiveLocation': archive_loc,
            'IsRestart': {'restart': restart,'Timestep': restart_timestep,'InitObjectId': initial_run_oid}
        },
        'ScenarioTag': { 
                    'Name of actual of hypothetical experiment': expt,
                    'scenario_runid': scenario_runid,
            },
        'CodeTag': { 
                'sim_type': sim_type,
                'IsLinear': linear,
                'quasi_linear': quasiLinear,
                'Has1DFluxes': has_1dflux,
                'git_hash': git_hash,
                'platform': platform,
                'execution_date': ex_date,
                'workflow_type': workflow_type
            },
        'Publications': { 
                'doi': [] 
            }
    }
    
    return metadata

def get_omega(out_dir, suffix):
    try:
        with open(os.path.join(out_dir, 'omega'+suffix)) as f:
            val = f.read().split()
            
        val = [float(v) for v in val] 
        if len(val) < 3:
            val = val + [np.nan for _ in range(3-len(val))]
    
    except:
        print('Omega file not found. Fill with NaN')
        val = [np.nan for _ in range(3)]
        
    return val
        

def get_time_for_diag(run_suffix):
    option = input('Please enter the tspan information for {}\n 1: Type them in manually.\n 2: Use default settings.\n 3. Use default settings for rest.\n'.format(run_suffix))
    if option == '1':      
        tspan = input('Please type start time and end time, separated by comma.\n').split(',')
        tspan[0] = float(tspan[0])
        tspan[1] = float(tspan[1])
    elif option == '2':
        tspan = None
    else:
        tspan = -1
    
    return tspan

def get_diag_with_user_input(out_dir, suffix,  manual_time_flag):

    if manual_time_flag:
        tspan = get_time_for_diag(suffix)
        if tspan == -1:
            manual_time_flag = False
            Diag_dict = get_diag_from_run(out_dir, suffix, None)
        else:
            Diag_dict = get_diag_from_run(out_dir, suffix, tspan) 
    else:
        Diag_dict = get_diag_from_run(out_dir, suffix, None)
        
    return Diag_dict, manual_time_flag

def get_data(key, *args):
    '''
    Use to get data from default files with functions defined in func_dic
    '''
    return func_dic[key](*args)

def get_data_by_func(user_func, *args):
    '''
    user_func takes args and should return a dictionary having at least two keys: '_header_' and '_data_'
    an example is provided as below: get_data_from_energy()
    '''
    return user_func(*args)

def get_data_from_energy(db, filepath):
    '''
    read GENE energy output, parsed into header and datapart
    '''
    fs = gridfs.GridFS(db)
    if fs.exists({"filepath": filepath}):
        file = fs.find_one({"filepath": filepath}) # assuming only one
        contents = file.read().decode('utf8').split('\n')
        header = []
        data = []
        for line in contents:
            if '#' in line:
                header.append(line)
            else:
                d_str = line.split()
                if d_str:
                    data.append([float(num) for num in d_str])
        
#        data = np.array(data)
        return {'_header_': header[:-1], '_data_': np.array(data)}
    
    else:
        print("No entry in current database matches the specified filepath.")
        return None

def get_data_from_nrg(db, filepath):
    fs = gridfs.GridFS(db)
    if fs.exists({"filepath": filepath}):
        file = fs.find_one({"filepath": filepath}) # assuming only one
        contents = file.read().decode('utf8').split('\n')
        header = []
        data = []
        time = []
        count = 0
        for line in contents[:-1]: # last line is ''
            if count % 2 == 0:
#               print(count)
               time.append(float(line))
            else:
                d_str = line.split()
                if d_str:
                    data.append([float(num) for num in d_str])
            count += 1
        
#        data = np.array(data)
        return {'_header_': header, '_time_': np.array(time), '_data_': np.array(data)}
    
    else:
        print("No entry in current database matches the specified filepath.")
        return None

def isfloat(a):
    try:
        float(a)
        return True
    except ValueError:
        return False

def to_float(a):
    try:
        b = float(a)
    except ValueError:
        b = a
    return b

def get_data_from_parameters(db, filepath):
    fs = gridfs.GridFS(db)
    if fs.exists({"filepath": filepath}):
        file = fs.find_one({"filepath": filepath}) # assuming only one
        contents = file.read().decode('utf8').split('\n')
        summary_dict=dict()
        for line in contents:
            if '&' in line:
                category = line[1:]
                sub_dict = {}
            elif '=' in line:
                pars = line.split('=')
                sub_dict[pars[0].rstrip()] = to_float(pars[1]) 
            elif '/' in line:
                summary_dict[category] = sub_dict            
            else:
                continue
            
        return summary_dict
    
    else:
        print("No entry in current database matches the specified filepath.")
        return None
  

def get_data_from_tracer_efit(db, filepath):      
    fs = gridfs.GridFS(db)
    if fs.exists({"filepath": filepath}):
        file = fs.find_one({"filepath": filepath}) # assuming only one
        contents = file.read().decode('utf8').split('\n')
        header_dict = {}
        data = []
        for line in contents:
            if '=' in line:
                item = line.split('=')
#                if '_' in item[1] or ' \' ' in item[1]:
                if isfloat(item[1]):
                    header_dict[item[0]] = float(item[1])
                else:
                    header_dict[item[0]] = item[1]
                    
            elif '/' in line or '&' in line:
                continue

            else:
                d_str = line.split()
                if d_str:
                    data.append([float(num) for num in d_str])
        
#        data = np.array(data)
        return {'_header_': header_dict, '_data_': np.array(data)}
    
    else:
        print("No entry in current database matches the specified filepath.")
        return None
    
func_dic = {'energy': get_data_from_energy,
            'nrg': get_data_from_nrg,
            'parameters': get_data_from_parameters
            }        

def get_file_list(out_dir, begin):
    '''
    Get files from out_dir that begins with "begin"
    '''
    files_list = []
    
    #unwanted filetype suffixes for general list
    bad_ext = ('.ps','.png', '.jpg', '.dat~', '.h5')
    
#    print('Searching in {} with key {}'.format(out_dir, begin))
    #scan files in GENE output directory, ignoring files in '/in_par', and return list
    
#    files = next(os.walk(out_dir))[2]
    files = os.listdir(out_dir)
    for count, name in enumerate(files, start=0):
        if name.startswith(begin) and name.endswith(bad_ext) == False: #and not os.path.isdir('in_par'):
            file = os.path.join(out_dir, name)
            if file not in  files_list:
                files_list.append(file)
            
    # print('{} files found in {} beginning with {}.'.format(len(files_list), out_dir, begin) )
    return files_list     


# def get_file_list(out_dir,fname):


def gridfs_put(db, filepath, sim_type):

    fs = gridfs.GridFS(db)
    with open(filepath, 'rb') as file:
        dbfile = fs.put(file, encoding='UTF-8', 
                        filepath=filepath,
                        filename=os.path.basename(filepath),
                        simulation_type=sim_type,
                        metadata=None)

    return dbfile
    
def gridfs_read(db, query):
    fs = gridfs.GridFS(db)
    file = fs.find_one(query)
    contents = file.read()
    return(contents)

def Array2Dict_dim1(npArray, key_names=None):
    '''
    Convert a 1d numpy array to dictionary
    '''
    assert len(npArray.shape) == 1, "Dimension of input numpy array should be 1."
    
    arraydict = dict()
    
    if key_names:
        for i in range(len(npArray)):
            arraydict[key_names[i]] = npArray[i]
    
    else:
        for i in range(len(npArray)):
            arraydict[str(i)] = npArray[i]
    
    return arraydict

def Array2Dict_dim2(npArray, row_keys=None, col_keys=None):
    '''
    Convert a 2d numpy array to dictionary
    '''
    assert len(npArray.shape) == 2, "Dimension of input numpy array should be 2."
    
    arraydict = dict()
    
    nrows, ncols = np.shape(npArray)
    if row_keys and col_keys:
        for i in range(nrows):
            row_dict = {}
            for j in range(ncols):
                row_dict[col_keys[j]] = npArray[i,j]
            arraydict[row_keys[i]] = row_dict
    
    else:
        for i in range(nrows):
            row_dict = {}
            for j in range(ncols):
                row_dict[str(j)] = npArray[i,j]
            arraydict[str(i)] = row_dict
        
    return arraydict

def Rep_OID(dic):
    '''
    Check a dictionary tree and replace any 'ObjectId' string to ObjectId object
    '''
    for key, val in dic.items():
        if isinstance(val, str) and 'ObjectId' in val:
#            oid_str = val[8:-1]
            oid_str = val[val.find('(')+1: val.find(')')].strip()
            dic[key] = ObjectId(oid_str)

        elif isinstance(val, dict):
            dic[key] = Rep_OID(val)
    return dic

def Str2Query(s):
    '''
    Convert a string s to python dictionary for querying the database
    '''
    q_dict = json.loads(s)
    q_dict = Rep_OID(q_dict)
    
    return q_dict

def get_oid_from_query(db, collection, query):
    
    records_found = collection.find(query)
    
    oid_list = []
    
    for record in records_found:
        oid_list.append(record['_id'])
        
    return oid_list

def _npArray2Binary(npArray):
    """Utility method to turn an numpy array into a BSON Binary string.
    utilizes pickle protocol 2 (see http://www.python.org/dev/peps/pep-0307/
    for more details).

    Called by stashNPArrays.

    :param npArray: numpy array of arbitrary dimension
    :returns: BSON Binary object a pickled numpy array.
    """
    return Binary(pickle.dumps(npArray, protocol=2), subtype=128 )

def _binary2npArray(binary):
    """Utility method to turn a a pickled numpy array string back into
    a numpy array.

    Called by loadNPArrays, and thus by loadFullData and loadFullExperiment.

    :param binary: BSON Binary object a pickled numpy array.
    :returns: numpy array of arbitrary dimension
    """
    return pickle.loads(binary)

def gridfs_put_npArray(db, value, filepath, filename, sim_type):
    '''
    Write numpy array to file and then to DB
    '''
    MAX_FILE_SIZE = 10 * 1024 * 1024  # 10 MB limit

    fs = gridfs.GridFS(db)
    binary_data = _npArray2Binary(value)
    data_size = len(binary_data)

    if data_size > MAX_FILE_SIZE: ## Ensure file is not too big
        print(f"Binary data for '{filename}' has size {data_size / (1024 * 1024)} MB. This exceeds the size limit of {MAX_FILE_SIZE / (1024 * 1024)} MB")
        print(f"Ignoring upload of this diagnostic: {filename}")
        obj_id = None
    else: 
        obj_id=fs.put(binary_data,encoding='UTF-8',
                    filename = filename,
                    simulation_type = sim_type,
                    filepath = filepath)
    
    return obj_id  
    
    
def load(db, collection, query, projection={'Metadata':1, 'gyrokineticsIMAS':1, 'Diagnostics':1}, getarrays=True):
    """Preforms a search using the presented query. For examples, see:
    See http://api.mongodb.org/python/2.0/tutorial.html
    The basic idea is to send in a dictionaries which key-value pairs like
    mdb.load({'basename':'ag022012'}).

    :param query: dictionary of key-value pairs to use for querying the mongodb
    :returns: List of full documents from the collection
    """
    
    results = collection.find(query, projection)
    
    if getarrays:
        allResults = [_loadNPArrays(db, doc) for doc in results]
    else:
        allResults = [doc for doc in results]
    
    if allResults:
#        if len(allResults) > 1:
#            return allResults
#        elif len(allResults) == 1:
#            return allResults[0]
#        else:
#            return None
        return allResults
    else:
        return None
    
def _loadNPArrays(db, document):
    """Utility method to recurse through a document and gather all ObjectIds and
    replace them one by one with their corresponding data from the gridFS collection

    Skips any entries with a key of '_id'.

    Note that it modifies the document in place.

    :param document: dictionary like-document, storable in mongodb
    :returns: document: dictionary like-document, storable in mongodb
    """
    fs = gridfs.GridFS(db)
    for (key, value) in document.items():
        if isinstance(value, ObjectId) and key != '_id':
            document[key] = _binary2npArray(fs.get(value).read())
        elif isinstance(value, dict):
            document[key] = _loadNPArrays(db, value)
    return document

def query_plot(db, collection, query, projection = {'Metadata':1, 'Diagnostics':1}):
    data_list = load(db, collection, query, projection)
    print('{} records found.'.format(len(data_list)))
    
    data_to_plot = [diag_plot(da) for da in data_list]
    
    for i in range(len(data_to_plot)):
         data_to_plot.plot_all()    
    
    
def isLinear(folder_name, sim_type):
    linear = None

    #check file for 'nonlinear' value
    suffixes = get_suffixes(folder_name, sim_type)
    
    if len(suffixes):
        suffix = suffixes[0] #assuming all parameters files are of the same linear/nonlinear type
        print('Scanning parameters for deciding linear/Nonlinar.')
    else:
        suffix = ''

    if sim_type=='GENE': 
        fname = os.path.join(folder_name, 'parameters' + suffix)
        if os.path.isfile( fname ):
            par = Parameters()
            par.Read_Pars( fname )
            pars = par.pardict
            linear = not pars['nonlinear']
            return(linear)
            
        #check folder name for nonlin
        elif folder_name.find('nonlin') != -1:
            linear = False
            return(linear)
        
        #check folder name for linear
        elif folder_name.find('linear') != -1:
            linear = True 
            return(linear)

        else:
            assert linear is None, "Can not decide, please include linear/nonlin as the suffix of your data folder!"
        
    elif sim_type=='CGYRO':
        fname=os.path.join(folder_name, suffix, 'input.cgyro')
        assert os.path.isfile(fname),"File %s does not exist"%(fname)

        non_lin = None
        with open(fname,'r') as f:
            for line in f: 
                if line.split('=')[0].strip()=='NONLINEAR_FLAG':
                    non_lin = int(line.split('=')[1].strip()) # Remove blank space to just get 0 or 1
        
        assert non_lin is not None, "Didn't find NONLINEAR_FLAG in file(%s)"%(fname) 

        linear = False if non_lin else True
        return linear
    
    elif sim_type=='TGLF':

        fname = os.path.join(folder_name, suffix, 'input.tglf')
        assert os.path.isfile(fname),"File %s does not exist"%(fname)

        with open(fname,'r') as f:
            for line in f: 
                val = line.split('=')

                if val[0].strip()=='USE_TRANSPORT_MODEL':
                    true_present  = [strg in val[1].strip() for strg in ['true','T','t']]
                    false_present = [strg in val[1].strip() for strg in ['false','F','f']]
                    if any(true_present): 
                        linear = False ## run is non linear
                    elif any(false_present):
                        linear = True 
                    else : 
                        print("Unknown entry in parameter file for field \"USE_TRANSPORT_MODEL\" ",line)
                        raise SystemError
                    break
        return linear
    
    elif sim_type=='GX':

        in_files = [f for f in os.listdir(os.path.join(folder_name,suffix)) if f.endswith('.in')]
        
        if len(in_files) > 1:
            print("Expected exactly one .in file in the folder, found %d. Using first one."%(in_files))
        fname=os.path.join(folder_name,suffix,in_files[0])
        assert os.path.isfile(fname),"File %s does not exist"%(fname)

        with open(fname,'r') as f:
            for line in f:
                strg = line.strip().split('#')[0].split('=')
                if (len(strg) == 2 and strg[0].strip() == 'nonlinear_mode'):
                    val = strg[1].strip()
                    if val in ['true','True','t','T']:
                        linear = False
                    elif val in ['false','False','f','F']:
                        linear = True
                    else : 
                        print("Unknown entry in parameter file for field \"nonlinear_mode\" ",line)
                        raise SystemError
                    break
        return linear
    
    elif sim_type=='GS2': # Only support linear GS2 for now
        return True

def isUploaded(out_dir,runs_coll):
    '''
    check if out_dir appears in the database collection.
    Assuming out_dir will appear no more than once in the database
    '''
    inDb = runs_coll.find({ "Metadata.DBtag.run_collection_name": out_dir })

    entries = list(inDb)
    uploaded = True if len(entries)>0 else False 
    
    # uploaded = False
    # for run in inDb:
    #     if run["Metadata"]["run_collection_name"] == out_dir: # seems redundent?
    #         uploaded = True
    #         break
    
    return uploaded

def not_uploaded_list(out_dir, runs_coll, write_to = None):
    '''
    Get all subfolders in out_dir that are not in the database yet
    '''
    not_uploaded = []
    for dirpath, dirnames, files in os.walk(out_dir):
        if str(dirpath).find('in_par') == -1 and str(files).find('parameters') != -1:
            if not isUploaded(dirpath, runs_coll):
                not_uploaded.append(dirpath)
    
    if write_to is not None and len(not_uploaded):
        with open(os.path.abspath(write_to),'w') as f:
            f.write('\n'.join(not_uploaded))
    
    return not_uploaded

def get_record(out_dir, runs_coll):
    '''
    Get a list of summary dictionary for 'out_dir' in the database
    '''
    inDb = runs_coll.find({ "Metadata.DBtag.run_collection_name": out_dir })
    record = []
    for run in inDb:
#        dic = dict()
#        for key, val in run.items():
#            dic[key] = val
#        record.append(dic)
        record.append(run)
    return record
   
def f_check_id_exists(db, _id):
    ''' Given an object ID, check if it exists in linear or nonlinear collections
    '''

    for collection in ['LinearRuns','NonlinRuns']:
        runs_coll =  getattr(db,collection)

        try: 
            record = runs_coll.find_one({"_id": _id})
        except Exception as e:
            # print(e)
            print("Invalid object ID",_id)
            return False
    
        if record is not None: return True

    print("Entry %s not found in database, please double check the id"%(_id))
    return False

def f_get_linked_oid(database, linked_id_string):
    '''
    Get linked ObjectID
    '''

    if linked_id_string is not None: 
        oid = ObjectId(linked_id_string)
        id_exists = f_check_id_exists(database, oid)

        if id_exists:
            print("Linked OID %s"%(oid))
            return oid
        else :
            return None

def download_file_by_path(db, filepath, destination, revision=-1, session=None):
    '''
    db: database name
    filepath: filepath stored in database, that is "db.fs.files['filepath']"
    destination: local path to put the file
    
    Attention: filename may correspond to multiple entries in the database
    '''
    fs = gridfs.GridFSBucket(db)
    records = db.fs.files.find({"filepath": filepath})
    count = 0
    for record in records:
        _id = record['_id']
        filename = record['filepath'].split('/')[-1]
        with open(os.path.join(destination, filename+'_mgk{}'.format(count) ),'wb+') as f:
            fs.download_to_stream(_id, f)
            count +=1
#            fs.download_to_stream_by_name(filename, f, revision, session)
        
    print("Download completed! Downloaded: {}".format(count))
    
def download_file_by_id(db, _id, destination, fname=None, session = None):
    '''
    db: database name
    _id: object_id
    destination: local path to put the file
    fname: name you want to call for the downloaded file
    '''

    fs = gridfs.GridFSBucket(db)
    if not fname:
        fname = db.fs.files.find_one(_id)['filename']
    if not os.path.exists(destination):
        Path(destination).mkdir(parents=True) 
    with open(os.path.join(destination, fname),'wb+') as f:   
        fs.download_to_stream(_id, f)
    print("Download completed!")
    
def download_dir_by_name(db, runs_coll, dir_name, destination):  
    '''
    db: database name
    dir_name: as appear in db.Metadata['run_collection_name']
    destination: destination to place files
    '''
    path = os.path.join(destination, dir_name.split('/')[-1])
    if not os.path.exists(path):    
        try:
            #os.mkdir(path)
            Path(path).mkdir(parents=True) 
        except OSError:
            print ("Creation of the directory %s failed" % path)
    #else:
    fs = gridfs.GridFSBucket(db)
    inDb = runs_coll.find({ "Metadata.DBtag.run_collection_name": dir_name })

    if 'generr' in inDb[0]['Files'].keys(): ## Fix for when 'generr' doesn't exist
        if inDb[0]['Files']['geneerr'] != 'None':    
            with open(os.path.join(path, 'geneerr.log'),'wb+') as f:
                fs.download_to_stream(inDb[0]['Files']['geneerr'], f, session=None)

    for record in inDb:
        '''
        Download 'files'
        '''
        for key, val in record['Files'].items():
            if val != 'None' and key not in ['geneerr']:
                filename = db.fs.files.find_one(val)['filename']
                with open(os.path.join(path, filename),'wb+') as f:
#                    fs.download_to_stream_by_name(filename, f, revision=-1, session=None)
                    fs.download_to_stream(val, f, session=None)
                record['Files'][key] = str(val)
        if 'generr' in record['Files'].keys():  ## Fix for when 'generr' doesn't exist 
            record['Files']['geneerr'] = str(record['Files']['geneerr'])
        
        '''
        Deal with diagnostic data
        '''
        diag_dict={}
        fsf=gridfs.GridFS(db)
        for key, val in record['Diagnostics'].items():
            if isinstance(val, ObjectId):
#                data = _loadNPArrays(db, val)
#                data = _binary2npArray(fsf.get(val).read()) # no need to store data
                record['Diagnostics'][key] = str(val)
#                data = _binary2npArray(fsf.get(val).read()) 
#                np.save( os.path.join(path,str(record['_id'])+'-'+key), data)
                diag_dict[key] = _binary2npArray(fsf.get(val).read())
            
        with open(os.path.join(path,str(record['_id'])+'-'+'diagnostics.pkl'), 'wb') as handle:
            pickle.dump(diag_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)
                
        record['_id'] = str(record['_id'])
        with open(os.path.join(path, 'mgkdb_summary_for_run'+record['Metadata']['run_suffix']+'.json'), 'w') as f:
            json.dump(record, f)
           
    print ("Successfully downloaded to the directory %s " % path)


def download_runs_by_id(db, runs_coll, _id, destination):
    '''
    Download all files in collections by the id of the summary dictionary.
    '''
    
    fs = gridfs.GridFSBucket(db)
    record = runs_coll.find_one({ "_id": _id })
    try:
        dir_name = record['Metadata']['DBtag']['run_collection_name']
    except TypeError:
        print("Entry not found in database, please double check the id")
        raise SystemExit
        
    path = os.path.join(destination, dir_name.split('/')[-1])

    if not os.path.exists(path):
        try:
#            path = os.path.join(destination, dir_name.split('/')[-1])
            #os.mkdir(path)
            Path(path).mkdir(parents=True)
        except OSError:
            print ("Creation of the directory %s failed" % path)
    #else:
    '''
    Download 'files'
    '''
    for key, val in record['Files'].items():
        if val != 'None':
            filename = db.fs.files.find_one(val)['filename']
            #print(db.fs.files.find_one(val)).keys()
            with open(os.path.join(path, filename),'wb+') as f:
#                    fs.download_to_stream_by_name(filename, f, revision=-1, session=None)
                fs.download_to_stream(val, f, session=None)
            record['Files'][key] = str(val)
            
    '''
    Deal with diagnostic data
    '''
    fsf=gridfs.GridFS(db)
    diag_dict = {}
    for key, val in record['Diagnostics'].items():
        if isinstance(val, ObjectId):
#                data = _binary2npArray(fsf.get(val).read()) # no need to store data
            record['Diagnostics'][key] = str(val)
#            data = _binary2npArray(fsf.get(val).read()) 
#            np.save( os.path.join(path,str(record['_id'])+'-'+key), data)
            diag_dict[key] = _binary2npArray(fsf.get(val).read())
            
    with open(os.path.join(path,str(record['_id'])+'-'+'diagnostics.pkl'), 'wb') as handle:
        pickle.dump(diag_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)        

    #print(record)
    record['_id'] = str(_id)

    with open(os.path.join(path, 'mgkdb_summary_for_run'+record['Metadata']['DBtag']['run_suffix']+'.json'), 'w') as f:
        json.dump(record, f)
    print("Successfully downloaded files in the collection {} to directory {}".format( record['_id'],path) )   
    
def update_mongo(db, metadata, out_dir, runs_coll, linear, suffixes=None):

    '''
    only update file related entries, no comparison made before update
    '''
    
    sim_type = metadata['CodeTag']['sim_type']

    fs = gridfs.GridFS(db)
    if suffixes is None:
        suffixes = get_suffixes(out_dir, sim_type)  
        
    update_option = input('Enter options for update:\n 0: Files shared by all runs, usually do not have a suffix. \n 1: Unique files used per run. Specify the keywords and suffixes. \n ')
    if update_option == '0':
        files_to_update = input('Please type FULL file names to update, separated by comma.\n').split(',')
        keys_to_update = input('Please type key names for each file you typed, separated by comma.\n').split(',')

        updated = []
        print('Uploading files .......')
        # update the storage chunk
        for doc, field in zip(files_to_update, keys_to_update):
            
            file = os.path.join(out_dir, doc)
            assert os.path.exists(file), "File %s not found"%(file)
            
            # delete ALL history
            grid_out = fs.find({'filepath': file})
            for grid in grid_out:
                print('File with path tag:\n{}\n'.format(grid.filepath) )
                fs.delete(grid._id)
                print('deleted!')

            with open(file, 'rb') as f:
                _id = fs.put(f, encoding='UTF-8', filepath=file, filename=file.split('/')[-1])
            
            updated.append([field, _id])
        
        # update the summary dictionary  
        print('Updating Metadata')              
        for entry in updated:
            for suffix in suffixes:                    
                runs_coll.update_one({ "Metadata.DBtag.run_collection_name": out_dir, "Metadata.DBtag.run_suffix": suffix}, 
                                 {"$set":{'Files.'+entry[0]: entry[1], 
                                          "Metadata.DBtag.last_updated": strftime("%y%m%d-%H%M%S")}}
                                 )
        print("Update complete")
                
    elif update_option == '1':
        files_to_update = input('Please type filenames (without suffixes) for files to update, separated by comma.\n').split(',')
        print("suffixes availables are:{}".format(suffixes))
        runs_to_update = input('Please type which suffixes to update, separated by comma. If you need to update all runs, just hit ENTER. \n').split(',')      
        # affect_QoI = input('Will the file change QoIs/Diagnostics? (Y/N)')
        affect_QoI = True

#        updated = []
        # update the storage chunk
        print('Uploading files .......')
        if len(runs_to_update[0]) != 0:
            run_suffixes = runs_to_update
        else:
            run_suffixes = suffixes
        
        for doc in files_to_update:
            manual_time_flag = True
            for suffix in run_suffixes:
                if affect_QoI:
                    input_fname = f_get_input_fname(out_dir, suffix, sim_type)
                    GK_dict, quasi_linear = create_gk_dict_with_pyro(input_fname, sim_type)   

                    if sim_type in ['CGYRO','TGLF','GS2','GX']:
                        Diag_dict = {}
                    elif sim_type=='GENE': 
                        Diag_dict, manual_time_flag = get_diag_with_user_input(out_dir, suffix, manual_time_flag)

                    run = runs_coll.find_one({ "Metadata.DBtag.run_collection_name": out_dir, "Metadata.DBtag.run_suffix": suffix})
                    for key, val in run['Diagnostics'].items():
                        if val != 'None':
                            # print((key, val))
                            fs.delete(val)
                            # print('deleted!')

                    for key, val in Diag_dict.items():
                        Diag_dict[key] = gridfs_put_npArray(db, Diag_dict[key], out_dir, key, sim_type)

                    runs_coll.update_one({ "Metadata.DBtag.run_collection_name": out_dir, "Metadata.DBtag.run_suffix": suffix },
                            { "$set": {'gyrokineticsIMAS': GK_dict, 'Diagnostics':Diag_dict}}
                                 )

                file = os.path.join(out_dir, doc  + suffix)
                grid_out = fs.find({'filepath': file})
                for grid in grid_out:
                    print('File with path tag:\n{}\n'.format(grid.filepath) )
                    fs.delete(grid._id)
                    print('deleted!')
                
                with open(file, 'rb') as f:
                    _id = fs.put(f, encoding='UTF-8', filepath=file, filename=file.split('/')[-1])

                runs_coll.update_one({ "Metadata.DBtag.run_collection_name": out_dir, "Metadata.DBtag.run_suffix": suffix }, 
                                 { "$set": {'Files.'+ doc: _id, "Metadata.DBtag.last_updated": strftime("%y%m%d-%H%M%S")} }
                                 )
        print("Update complete")
    
    else:
        print('Invalid input. Update aborted.')
        pass
    
def remove_from_mongo(out_dir, db, runs_coll):
    #find all documents containing collection name
        
    inDb = runs_coll.find({ "Metadata.DBtag.run_collection_name": out_dir })        
    fs = gridfs.GridFS(db)
    for run in inDb:
        # delete the gridfs storage:
        for key, val in run['Files'].items():
#            print(val)
#            if (key in all_file_keys) and val != 'None':
##                print((key, val))
#                target_id = ObjectId(val)
#                print((key, target_id))
#                fs.delete(target_id)
#                print('deleted!')
            if val != 'None':
                print((key, val))
                fs.delete(val)
                print('deleted!')
#                if fs.exists(target_id):
#                    print("Deleting storage for entry \'{}\' deleted with id: {}").format(key, val)
#                    fs.delete(target_id)
#                    print("Deleted!")
                
        for key, val in run['Diagnostics'].items():
            if val != 'None':
                print((key, val))
                fs.delete(val)
                print('deleted!')
                
#        delete the header file
        runs_coll.delete_one(run)
        
def f_update_global_var(global_vars, out_dir, suffix, sim_type, is_linear, large_files, count):
    '''
    Update global variables for each suffix
    '''
    ## For GENE, add geometry file to required and main files 
    ## Also modify large files when encountering many species 

    if sim_type=='GENE': ## Add geometry file for GENE 
        par = Parameters()
        par_file = os.path.join(out_dir,'parameters'+suffix)
        par.Read_Pars(par_file)
        pars = par.pardict
        ## Get geometry from parameters file and add that to list of files to save
        if 'magn_geometry' in pars:
            geom_file = pars['magn_geometry'][1:-1]
            global_vars.required_files.append(geom_file)
            global_vars.Docs.append(geom_file)

        n_spec = pars['n_spec']
        
        if large_files:
            if 'name1' in pars and 'mom' in global_vars.Docs_L:
                # global_vars.Docs_L.pop(global_vars.Docs_L.index('mom'))
                global_vars.Docs_L.remove('mom')
                for i in range(n_spec): # adding all particle species
                    global_vars.Docs_L.append('mom_'+pars['name{}'.format(i+1)][1:-1])

        if not is_linear: ## No omega file needed for non-linear runs
            global_vars.required_files.remove('omega')
            # global_vars.Docs.remove('omega')

    if large_files: 
        global_vars.Docs +=global_vars.Docs_L
    global_vars.update_docs_keys()

def f_get_full_fname(sim_type, fldr, suffix, fname):
    '''
    Give full file name for sim_type using suffix and filename
    '''
    if sim_type=='GENE':
        full_fname = os.path.join(fldr,fname+suffix)
    else: 
        full_fname = os.path.join(fldr,suffix,fname)

    return full_fname

def upload_file_chunks(db, out_dir, sim_type, suffix=None, run_shared=None, global_vars=None):
    '''
    This function does the actual uploading of gridfs chunks and
    returns object_ids for the chunk. If a ValueError is raised due to file size,
    it returns the current state of the dictionaries along with an error flag.
    '''
    
    MAX_FILE_SIZE = 10 * 1024 * 1024  # 10 MB limit 

    _docs, _keys = global_vars.all_file_docs, global_vars.all_file_keys

    file_upload_dict = {}
    s_dict = {}
    error_occurred = False
    error_message = None

    try:
        for doc, key in zip(_docs, _keys):
            file = f_get_full_fname(sim_type, out_dir, suffix, doc)
            file_upload_dict[key] = {'full_fname': file, 'oid': None}

            if os.path.isfile(file):
                ## Ensure file is not too big
                file_size = os.path.getsize(file)

                if file_size > MAX_FILE_SIZE:
                    raise ValueError(
                        "Size of the file %s is %s MB and it exceeds size limit of %s MB" %
                        (file, file_size / (1024 * 1024), MAX_FILE_SIZE / (1024 * 1024))
                    )

                _id = gridfs_put(db, file, sim_type)
                file_upload_dict[key]['oid'] = _id
            else:
                print(f'{file} not found in {out_dir}')

        
        ### Add shared files 
        if isinstance(run_shared, list):
            for sh in run_shared:
                key = sh.replace('.', '_')
                file = os.path.join(out_dir, sh)
                s_dict[key] = {'full_fname': file, 'oid': None}

                if os.path.isfile(file):
                    ## Ensure file is not too big
                    file_size = os.path.getsize(file)

                    if file_size > MAX_FILE_SIZE:
                        raise ValueError(
                            "Size of the file %s is %s MB and it exceeds size limit of %s MB" %
                            (file, file_size / (1024 * 1024), MAX_FILE_SIZE / (1024 * 1024))
                        )

                    _id = gridfs_put(db, file, sim_type)
                    s_dict[key]['oid'] = _id
                else:
                    print(f'{file} not found in {out_dir}')

        return file_upload_dict, s_dict, error_occurred, error_message

    except ValueError as e:
        error_occurred = True
        error_message = str(e)
        return file_upload_dict, s_dict, error_occurred, error_message

def f_get_input_fname(out_dir, suffix, sim_type):
    ''''
    Get the name of the input file with suffix for the simluation type
    '''

    fname_dict = {'CGYRO':os.path.join(out_dir,suffix,'input.cgyro'),
                    'TGLF':os.path.join(out_dir,suffix,'input.tglf'),
                    'GENE':os.path.join(out_dir,'parameters{0}'.format(suffix)),
                    'GS2': os.path.join(out_dir,suffix,'gs2.in'),
                    'GX': os.path.join(out_dir,suffix,'gx.in')
                }

    return fname_dict[sim_type]

def upload_runs(db, metadata, out_dir, is_linear=True, suffixes=None, run_shared=None,
                large_files=False, verbose=True, manual_time_flag=True, global_vars=None):
    """
    Uploads simulation run data to the database, handling both linear and nonlinear runs.

    Parameters:
    - db: Database connection object.
    - metadata: Dictionary containing metadata for the run.
    - out_dir: Output directory containing simulation files.
    - is_linear: Boolean indicating if the run is linear (True) or nonlinear (False). Default: True.
    - suffixes: List of suffixes for files to upload. If None, determined automatically.
    - run_shared: List of shared files to upload (optional).
    - large_files: Boolean to handle large file uploads. Default: False.
    - verbose: Boolean to print detailed output. Default: True.
    - manual_time_flag: Boolean to handle user-specified time spans for diagnostics. Default: True.
    - global_vars: Object containing global variables for the upload process.

    Returns:
    None
    """
    sim_type = metadata['CodeTag']['sim_type']

    # Connect to the appropriate collection
    runs_coll = db.LinearRuns if is_linear else db.NonlinRuns

    # Update files dictionary
    if suffixes is None:
        suffixes = get_suffixes(out_dir, sim_type)

    if isinstance(run_shared, list):
        shared_not_uploaded = [True for _ in run_shared]
    else:
        shared_not_uploaded = [False]
    shared_file_dict = {}

    for count, suffix in enumerate(suffixes):
        try:
            print('='*40)
            print(f'Working on files with suffix: {suffix} in folder {out_dir}.......')

            f_update_global_var(global_vars, out_dir, suffix, sim_type, is_linear, large_files, count)

            uploaded_ids = {}

            files_exist = f_check_required_files(global_vars, out_dir, suffix, sim_type)
            assert files_exist, "Required files don't exist. Skipping folder"

            # Compute gyrokinetics IMAS using pyrokinetics package
            print("Computing gyrokinetics IMAS using pyrokinetics")
            input_fname = f_get_input_fname(out_dir, suffix, sim_type)
            GK_dict, quasi_linear = create_gk_dict_with_pyro(input_fname, sim_type)

            # Upload files to DB
            print('Uploading files ....')

            if count == 0:
                f_dict, s_dict,err_occured, err_msg = upload_file_chunks(db, out_dir, sim_type, suffix, run_shared, global_vars)
                shared_file_dict = {k: v['oid'] for k, v in s_dict.items()}
            else:
                f_dict, s_dict, err_occured, err_msg = upload_file_chunks(db, out_dir, sim_type, suffix, None, global_vars)

            files_dict = {k: v['oid'] for k, v in f_dict.items()}
            files_dict = {**files_dict, **shared_file_dict}
            uploaded_ids = {k:v for k,v in files_dict.items() if v is not None}
            if err_occured: 
                print('Error occured during input file upload')
                raise ValueError(err_msg)

            print('='*60)
            # Metadata dictionary
            time_upload = strftime("%y%m%d-%H%M%S")

            metadata['DBtag']['run_collection_name'] = out_dir
            metadata['DBtag']['run_suffix'] = '' + suffix
            metadata['DBtag']['time_uploaded'] = time_upload
            metadata['DBtag']['last_updated'] = time_upload
            metadata['CodeTag']['IsLinear'] = is_linear
            metadata['CodeTag']['quasi_linear'] = quasi_linear
            metadata['CodeTag']['Has1DFluxes'] = GK_dict['non_linear']['fluxes_1d']['particles_phi_potential'] != 0

            meta_dict = metadata

            # Handle diagnostics based on sim_type
            if sim_type in ['CGYRO', 'TGLF', 'GS2', 'GX']:
                Diag_dict = {}
            elif sim_type == 'GENE':
                print('='*60)
                # print('\n Working on diagnostics with user specified tspan .....\n')
                Diag_dict, manual_time_flag = get_diag_with_user_input(out_dir, suffix, manual_time_flag)
                print('='*60)

                if is_linear:
                    # Add omega info to Diag_dict for linear runs
                    omega_val = get_omega(out_dir, suffix)
                    Diag_dict['omega'] = {
                        'ky': omega_val[0],
                        'gamma': omega_val[1],
                        'omega': omega_val[2]
                    }

                for key, val in Diag_dict.items():
                    oid = gridfs_put_npArray(db, val, out_dir, key, sim_type)
                    Diag_dict[key] = oid ## Rewrite array with oid of stored file
                    if oid is not None: 
                        uploaded_ids[key] = oid

            # Combine dictionaries and upload
            run_data = {
                'Metadata': meta_dict,
                'Files': files_dict,
                'gyrokineticsIMAS': GK_dict,
                'Diagnostics': Diag_dict
            }
            
            main_record_oid = runs_coll.insert_one(run_data).inserted_id

            print(f'Files with suffix: {suffix} in folder {out_dir} uploaded successfully.')
            print('='*40)
            if verbose:
                print('A summary is generated as below:\n')
                print(run_data)

        except Exception as e1:
            print(e1)
            print(f"Skip suffix {suffix} in \n {out_dir} \n")
            global_vars.troubled_runs.append(out_dir + '##' + suffix)
            print('cleaning ......')
            fs = gridfs.GridFS(db)
            try:
                for key, _id in uploaded_ids.items():
                    fs.delete(_id)
                    print(f'{key}: {_id} deleted.')
            except Exception as e3:
                print(f"Error deleting files from gridfs with exception:\t {e3}")
                pass

        global_vars.reset_docs_keys(sim_type)


def upload_to_mongo(db, linear, metadata, out_dir, suffixes=None, run_shared=None,
                    large_files=False, verbose=False, manual_time_flag=False, global_vars=None, no_prompts=False, reupload_if_exists=False):
    """
    Wrapper function to upload simulation runs to MongoDB, handling both linear and nonlinear runs.

    Parameters:
    - db: Database connection object.
    - linear: Boolean indicating if the run is linear (True) or nonlinear (False).
    - metadata: Dictionary containing metadata for the run.
    - out_dir: Output directory containing simulation files.
    - suffixes: List of suffixes for files to upload. If None, determined automatically.
    - run_shared: List of shared files to upload (optional).
    - large_files: Boolean to handle large file uploads. Default: False.
    - verbose: Boolean to print detailed output. Default: True.
    - manual_time_flag: Boolean to handle user-specified time spans for diagnostics. Default: False.
    - global_vars: Object containing global variables for the upload process.
    - no_prompts: Autoupload with no prompts. Default = False
    - reupload_if_exists: Delete and reupload if existing folder name is present in DB. Default: False
    Returns:
    None
    """
    # Connect to the appropriate collection based on linear flag
    runs_coll = db.LinearRuns if linear else db.NonlinRuns

    # Determine run type for printing
    run_type = 'linear' if linear else 'nonlinear'
    print(f'Upload {run_type} runs ******')

    # Check if folder is already uploaded
    if isUploaded(out_dir, runs_coll):
        print(f'Folder tag:\n {out_dir} \n exists in database')
        
        if no_prompts: 
            update='0' if reupload_if_exists else '1'
        else: 
            update = input(f'You can:\n 0: Delete and reupload folder? \n 1: Run an update (if you have updated files to add) \n Press any other keys to skip this folder.\n')
        
        if update == '0':
            # Delete and reupload
            print("Deleting {out_dir} and reuploading")
            remove_from_mongo(out_dir, db, runs_coll)
            upload_runs(db, metadata, out_dir, is_linear=linear, suffixes=suffixes, run_shared=run_shared,
                        large_files=large_files, verbose=verbose, manual_time_flag=manual_time_flag, global_vars=global_vars)
        elif update == '1':
            update_mongo(db, metadata, out_dir, runs_coll, linear)
        else:
            print(f'Run collection \'{out_dir}\' skipped.')
    else:
        print(f'Folder tag:\n{out_dir}\n not detected, creating new.\n')
        upload_runs(db, metadata, out_dir, is_linear=linear, suffixes=suffixes, run_shared=run_shared,
                    large_files=large_files, verbose=verbose, manual_time_flag=manual_time_flag, global_vars=global_vars)



"""fsamom_data.py

Contains the infrastructure to handle fsamom_spec_fext files

"""
import numpy as np
from bisect import bisect_left
from .base_file import TimeSeries
from ..utils.ParIO import Parameters


class FSAmomFile(TimeSeries):
    """Class to handle a specific fsamom file

    :param filename: Filename, typically fsamom[n,p,e]_[species]_fext
    :param common: CommonData object of the run
    """

    def __init__(self, filename, common):
        super().__init__(filename, common)
        self.fsacolumns = 13  # Number of columns in fsamom file
        self.dataarray = []
        self.momtitle = []
        self.generate_timeseries()

    def generate_timeseries(self):
        """ Read flux surface averaged moments """
        blocks = []
        self.timearray = []
        print('Reading {}\n'.format(self.objectname))
        try:
            with open(self.objectname) as fsafile:
                next(fsafile)  # skip header
                for line in fsafile:
                    if not line or line.startswith('\n'):
                        continue
                    if line.startswith("#   x/a"):
                        continue
                    if line.startswith('#'):
                        self.timearray.append(float(line.split()[1]))
                        blocks.append([])
                    else:
                        blocks[-1].append(line)
        except IOError:
            print("IOError: probably fsa file does not exist: {}".format(self.objectname))
            raise
        self.timearray = np.array(self.timearray)
        self.check_times()
        pos = self.calc_positions()
        self._addfsadata(blocks, pos)
        self.timearray = self.timearray[pos]

    def _addfsadata(self, blocks, pos):
        for entry in pos:
            self.dataarray.append(np.empty((self.cm.pnt.nx0, self.fsacolumns)))
            for i in range(0, self.cm.pnt.nx0):
                line = blocks[entry][i]
                for var in range(self.fsacolumns):
                    try:
                        self.dataarray[-1][i, var] = float(line.split()[var])
                    except ValueError:
                        self.dataarray[-1][i, var] = 0.0


class FSAmomData(object):
    """Class to handle the flux surface averaged moment diag of a specific run

    Version that handles the new version of the diagnostic with three output files
    """

    def __init__(self, common):
        self.isdatapresent = False
        self.cm = common
        self.fsaspec = []

    def getfsadata(self):
        """Read the data from the fsa_moments diagnostic"""
        try:
            if self.cm.pnt.istep_fsa_moments == 0:
                print("No fsamom diagnostic included in GENE run")
                return
        except AttributeError:
            print("No fsamom diagnostic included in GENE run")
            return
        for n in range(0, self.cm.pnt.n_spec):
            filename = 'fsamomn_{}{}'.format(self.cm.specnames[n], self.cm.fileextension)
            self.fsaspec.append(FSAmomFile(filename, self.cm))
            self.fsaspec[-1].momtitle = r"Density"
            filename = 'fsamomp_{}{}'.format(self.cm.specnames[n], self.cm.fileextension)
            self.fsaspec.append(FSAmomFile(filename, self.cm))
            self.fsaspec[-1].momtitle = r"$v_\parallel$"
            filename = 'fsamome_{}{}'.format(self.cm.specnames[n], self.cm.fileextension)
            self.fsaspec.append(FSAmomFile(filename, self.cm))
            self.fsaspec[-1].momtitle = r"$v^2$"
        self.isdatapresent = True


class FSAmomData_legacy(object):
    """Class to handle the flux surface averaged moment diag of a specific run

    Version that handles the old version of the diagnostic with one output files
    """

    def __init__(self, common):
        self.isdatapresent = False
        self.cm = common
        self.fsaspec = []

    def getfsadata(self):
        """Read the data from the fsa_moments diagnostic"""
        try:
            if self.cm.pnt.istep_fsa_moments == 0:
                print("No fsamom diagnostic included in GENE run")
                return
        except AttributeError:
            print("No fsamom diagnostic included in GENE run")
            return
        for n in range(0, self.cm.pnt.n_spec):
            filename = 'fsamom_{}{}'.format(self.cm.specnames[n], self.cm.fileextension)
            self.fsaspec.append(FSAmomFile(filename, self.cm))
            self.fsaspec[-1].momtitle = r"Density, $v_\parallel$ or $v^2$"
        self.isdatapresent = True



""" Module containing functions for manipulating 1D timetraces"""
import numpy as np
import math

def autocorr_time(time, series, tau = 3):
    z = np.autocorr(series)
    return np.where(z<np.exp(-1) - time[0])


def mytrapz(yvar, timefld):
    """ Trapezoid rule, adjusted for single timestep

    Operates on the first dimension of yvar (typically time)
    hence the name timefld for the integration variable samples
    :param yvar: Variable to integrate over
    :param timefld: The integration variable (time)
    :returns: The integration result (dim(yvar)-1)
    """
    timefld = np.array(timefld)
    yvar = np.real_if_close(yvar)
    if timefld.size == 1:
        return np.atleast_2d(yvar)[0]
    else:
        if yvar.shape[0] != len(timefld):
            raise ValueError("First dimension of yvar and timefld do not match")
        tmpt = timefld/(timefld[-1] - timefld[0])
        return np.trapz(yvar, x=tmpt, axis=0)
    
    
def find_nearest(array, value):
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return array[idx-1], idx-1
    else:
        return array[idx], idx
    


"""plotfsamom.py: Module to do the plotting of flux-surface averaged moments """

import numpy as np
import ..utils.averages as av
from .baseplot import plt, Plotting


class PlotFSA(Plotting):
    """ Class to plot time averaged FSA moment data

    :param fsalist: List of FSAmomdata objects to plot
    """

    def __init__(self, fsalist):
        super().__init__()
        self.fsalist = fsalist
        self.figlist = []
        self.axlist = []
        self.leglist = []
        self.labellist = ["x", "tot", r"$d/dt_{tot}$", "coll", r"$d\chi/dxy$", r"$f_1 src$",
                          r"$f_1 buff$", r"$f_0 term$", r"$v_D f_1$", r"hyp_v", r"hyp_z", r"hyp_x",
                          r"zv poisson"]

    def createfigs(self, poutput):
        """ Generate plot windows

        :param poutput: Generate the plots as pdfs as well
        """
        nfigs = len(self.fsalist[0].fsaspec)*len(self.fsalist)
        nfigsperfsa = len(self.fsalist[0].fsaspec)
        self.figlist = [plt.figure() for _ in range(nfigs)]
        # One plot window for each moment, species and quantity, one axis object in each
        self.axlist = [fig.add_subplot(111) for fig in self.figlist]
        for ifsa, fsa in enumerate(self.fsalist):
            self.plot_fsa(self.axlist[ifsa*nfigsperfsa:(ifsa+1)*nfigsperfsa], fsa)
        if poutput:
            for ifig, fig in enumerate(self.figlist):
                fig.savefig("fsamoments{}.pdf".format(ifig))

    def plot_fsa(self, axlist, fsa):
        """ Plot the results of the FSAmom diagnostic

        :param axlist: List of axes objects (one per each species)
        :param fsa: FSAmomData object containing the data
        """
        if not fsa.isdatapresent:
            print("No fsa data to plot")
            return
        Afs = (4*np.pi ** 2*self.fsalist[0].cm.pnt.major_R*self.fsalist[0].cm.pnt.minor_r /
               self.fsalist[0].cm.pnt.rhostar)
        intfsa = np.zeros((fsa.cm.pnt.nx0, fsa.fsaspec[0].fsacolumns))
        for n, fsd in enumerate(fsa.fsaspec):
            avfsa = av.mytrapz(fsd.dataarray, fsd.timearray)
            xs = fsd.dataarray[0][:, 0]
            for i in range(fsa.cm.pnt.nx0 - 1):
                intfsa[i+1, :] = intfsa[i, :] + avfsa[i+1, :]*Afs*xs[i+1]*(xs[1] - xs[0])
            intfsa = avfsa
            for iterm in range(2, 5):
                axlist[n].plot(xs, intfsa[:, iterm], label=self.labellist[iterm])
            axlist[n].legend()
            axlist[n].set_xlabel(r'$x/a$')
            axlist[n].set_title(fsd.momtitle)



""" Base classes for file reading"""

import struct
from bisect import bisect_left, bisect_right
import os
from os.path import getsize
import abc
import h5py
import numpy as np
import types


class File(abc.ABC):
    """ Base class to read files from GENE runs"""

    @abc.abstractmethod
    def __init__(self, folder=None, extension=None, file_type=None, parameters=None, spec=None):
        """ call this to initialize the file"""
        
    @abc.abstractmethod
    def redirect(self, folder=None, extension=None, parameters=None):
        """ Call this routine to read from a new file of the same type"""

    @abc.abstractmethod
    def get_filename(self):
        """ Construct the file name from information present in the object """

    @abc.abstractmethod
    def get_timearray(self):
        """ Fetch the time stamps from the file""" 
        
    def get_minmaxtime(self):
        """ Return the first and last time stamp in the file
        """
        if not self.timearray:
            self.get_timearray()
        return self.timearray[0], self.timearray[-1]           
        
    def get_times_and_inds(self):
        """ this is to be used for collecting all times in the file
            will return times, steps, and file extension.
            """
        if not self.timearray:
            self.__reset_tinds()
            time_list = self.get_timearray()

        times = np.asarray(self.timearray)
        steps = np.arange(0, times.size, 1)
        file_list = [self.extension]*times.size

        return times, steps, file_list

    def _find_nearest_time(self, time):
        """ Convert an time to its nearest step """
        pos = bisect_left(self.timearray, time)
        if pos == 0:
            return self.timearray[0]
        if pos == len(self.timearray):
            return self.timearray[-1]
        before = self.timearray[pos - 1]
        after = self.timearray[pos]
        if after - time < time - before:
            return after
        else:
            return before
        
    def _time_to_step(self, idx):
        """ Convert an exact time to its step """
        if not self.timearray:
            self.get_timearray()
        self.loaded_step[idx]=np.where(self.timearray == self.loaded_time[idx])
        
    def _step_to_time(self, idx):
        """ Convert the step to its time """
        if not self.timearray:
            self.get_timearray()
        self.loaded_time[idx]=self.timearray[self.loaded_step[idx]]    
        
    def _fix_step(self, step):
        """ Allow for negative steps like python """
        if step<0 :
            if not self.timearray:
                self.get_timearray()
            step=np.arange(len(self.timearray))[step]  
        return step
    
    def __reset_tinds(self):
        """ Reset time indices """
        self.loaded_time = [None]*self._nfields
        self.loaded_step = [None]*self._nfields
        self.timearray = []

    @abc.abstractmethod
    def _readvar(self, *args):
        """ Fetch a specific variable at the set time from the file"""

    def get_var(self, name):
        """ Fetch the data at set time, only accessing file if necessary """
        varidx = {v: k for k, v in self.varname.items()}
        if not self.loaded_time[varidx[name]] == self.time:
            self.loaded_time[varidx[name]] = self.time
            setattr(self, name + "_data", self._readvar(varidx[name]))
        return getattr(self, name + "_data")

    def _define_variables(self, parameters):
        """ provided a file type (e.g. field) it will return the list of variables
            that are in that file"""

        VARS_TO_FILE_MAP = {'field': {0: 'phi', 
                                      1: 'A_par', 
                                      2: 'B_par'},
                            'mom':  {0: "dens",
                                     1: "T_par", 
                                     2: "T_perp",
                                     3: "q_par", 
                                     4: "q_perp", 
                                     5: "u_par",
                                     6: 'densI1', 
                                     7: 'TparI1', 
                                     8: 'TppI1'},
                            'srcmom':  {0: "ck_heat_M00", 
                                        1: "ck_heat_M10",
                                        2: "ck_heat_M22",
                                        3: "ck_part_M00",
                                        4: "ck_part_M10",
                                        5: "ck_part_M22",
                                        6: "f0_term_M00",
                                        7: "f0_term_M10",
                                        8: "f0_term_M22"},
                            'vsp':  {0: '<f_>',
                                     1: 'G_es', 
                                     2: 'G_em', 
                                     3: 'Q_es',
                                     4: 'Q_em'}}                       
        return dict(list(VARS_TO_FILE_MAP[self.file_type].items())[0: self._set_nfields(parameters)]) 


    def _set_nfields(self, parameters):
        NFIELDS_TO_FILE_MAP = {'field': int(parameters.pardict['n_fields']),
                               'mom': int(parameters.pardict['n_moms']),
                               'srcmom': 9,
                               'vsp': 5}
        return NFIELDS_TO_FILE_MAP[self.file_type]  

    def _is_complex(self):
        IS_COMPLEX_MAP = {'field' : True, 
                          'mom'   : True, 
                          'srcmom': False,
                          'vsp'   : False}

        return IS_COMPLEX_MAP[self.file_type]

    def _set_boxsize(self, parameters):
        BOX_SIZE_MAP = {'field' : np.array([parameters.pnt.nx0, parameters.pnt.nky0, parameters.pnt.nz0]),
                        'mom'   : np.array([parameters.pnt.nx0, parameters.pnt.nky0, parameters.pnt.nz0]),
                        'src'   : np.array([parameters.pnt.nx0]),
                        'vsp'   : np.array([parameters.pnt.nz0, parameters.pnt.nv0, parameters.pnt.nw0])}

        return BOX_SIZE_MAP[self.file_type]

    def _set_prepath(self, file_type, spec):
        PREPATH_TO_FILE_MAP = {'field': '/field/',
                               'mom': '/mom_{}/'.format(spec),
                               'srcmom': '/srcmom_{}/'.format(spec),
                               'vsp': '/vsp/'}

        return PREPATH_TO_FILE_MAP[file_type]

    def _check_for_trap_split(self, dic, run_data):
        # TODO
        return dic

class BinaryFile(File):
    """ Base class to read Fortran binary (unformatted) files from GENE runs        
    """
    def __init__(self, folder=None, extension=None, file_type=None, parameters=None, spec=None):
        super().__init__(folder=None, extension=None, file_type=None, parameters=None, spec=None)
        
        # these are public so we can reuse the object if we have to by redirecting it to a new folder/extension
        self.folder = parameters.in_folder                   # current folder we are pointing at
        self.extension = parameters.extension                # current extension we are pointing at
        self.file_type=file_type
        self.__spec=spec
        self.varnames = None
        
        # fid for reading
        self.__fid = None                                    # file identifier for reading 
   
        # point the object to folder/file_type+extension, setting fid, variables and sizes.
        self.redirect(folder, extension, parameters)      
    
        self.timearray = []                                 # time series saved in the file      
 
    def get_filename(self):          
        rstring = self.file_type
        if self.__spec:
            rstring += '_' + self.__spec
        rstring+=self.extension
        
        return os.path.join(self.folder, rstring)
            
    def redirect(self, folder=None, extension=None, parameters=None):
        """ Call this routine to read from a new file. To allow for precision change
            must reset all size."""
        
        if folder == self.folder and self.extension==extension:
            #already athe right file
            pass
        
        if folder:
            self.folder = folder
        if extension:
            self.extension = extension
            
        # make sure the file is closed
        try:
            self.__fid.close()
        except (AttributeError, OSError):
            pass        
        
        # get the filename we want
        self.filename = self.get_filename()
        
        # now oen the file
        try:
            self.__fid = open(self.filename, 'rb')
        except:
            raise Exception(self.filename + " does not exist")
        
        # now set the content of the file
        # do it only if is the first time
        if not self.varnames:
            self.varnames = self._define_variables(parameters)
        
            # data size
            self.boxsize = self._set_boxsize(parameters)
            self._nfields = self._set_nfields(parameters)
        
            # data properties for reading
            self.__set_datatypes(parameters)
            self.__set_sizes(parameters)
            self.__tentry, self.__tesize = self.__set_time_entry(parameters)
        
        #reset times and ind
        self.timearray = [] 
        self.loaded_time = [None]*self._nfields              # time step currently loaded for a given variable
        self.loaded_step = [None]*self._nfields              # timestep at which fid is pointing       


    def __set_datatypes(self, parameters):
        if parameters.pardict['PRECISION'] == 'DOUBLE':
            self.__nprt = np.dtype(np.float64)
            self.__npct = np.dtype(np.complex128)
        else:
            self.__nprt = np.dtype(np.float32)
            self.__npct = np.dtype(np.complex64)
            
        if  parameters.pardict['ENDIANNESS'] == 'BIG':
            self.__nprt=self.__nprt.newbitorder()
            self.__npct=self.__npct.newbitorder()
                

    def __set_sizes(self, parameters):
        """we must set based on current file precision"""
        self.__intsize = 4
        self.__realsize = 8 if parameters.pardict['PRECISION'] == 'DOUBLE' \
                        else 4
        self.__complexsize = 2*self.__realsize
        
        if self._is_complex():
            self.__entrysize = np.prod(self.boxsize) * self.__complexsize
        else:    
            self.__entrysize = np.prod(self.boxsize) * self.__realxsize   
        self.__leapfld = self._nfields*(self.__entrysize + 2*self.__intsize)

    def __set_time_entry(self,parameters):
        """ Defines the struct for a time entry """
        # Fortran writes records as sizeof(entry) entry sizeof(entry)
        if parameters.pardict['PRECISION'] == 'DOUBLE':
            if parameters.pardict['ENDIANNESS'] == 'BIG':
                timeentry = struct.Struct('>idi')
            else:
                timeentry = struct.Struct('=idi')
        else:
            if parameters.pardict['ENDIANNESS'] == 'BIG':
                timeentry = struct.Struct('>ifi')
            else:
                timeentry = struct.Struct('=ifi')

        return timeentry, timeentry.size

    def offset(self, idx):
        """Calculate offset in field file for a given self.time and variable"""
        if idx in range(self._nfields):
            return self.__tesize + self.loaded_step[idx]*(self.__tesize + self.__leapfld) + idx*(
                self.__entrysize + 2*self.__intsize) + self.__intsize
        else:
            print("Something went terribly wrong here")

    def _readvar(self, idx):
        """ Return 3d field data at the time set in self.time"""
        self.__fid.seek(self.offset(idx))       
        var3d = np.fromfile(self.__fid, count=np.prod(self.boxsize),
                            dtype=self.__npct)
        
        return var3d.reshape(tuple(self.boxsize), order="F")

    def get_timearray(self):
        """ Get time array from file """
        self.timearray = []
        self.__fid.seek(0)
        for _ in range(int(getsize(self.filename)/(self.__leapfld + self.__tesize))):
            self.timearray.append(float(self.__tentry.unpack(self.__fid.read(self.__tesize))[1]))
            self.__fid.seek(self.__leapfld, 1)
        return self.timearray   
    
def read_method(afile, name, idx):
    def __my_read_method(self, time=None, step=None, extension=None):
        if extension and not extension==self.extension:
            #this does not support change in precision
            self.redirect(self.folder, extension)
        
        if time is not None and not self.loaded_time[idx] == time:
            # loading by time but the data is not what we have been asked
            self.loaded_time[idx] = time
            # convert to step
            self._time_to_step(idx)
            #load the step
            setattr(self, name + "_data", self._readvar(idx))
        elif step is not None and not self.loaded_step[idx] == step:            
            # loading by step but the data is not what we have been asked
            self.loaded_step[idx] = self._fix_step(step)
            # convert to time
            self._step_to_time(idx)
            # load the step  
            setattr(self, name + "_data", self._readvar(idx))
        return getattr(self, name + "_data")
     
    setattr(afile, name, types.MethodType(__my_read_method, afile))
    

class H5File(File):
    """ Base class to read HDF5 files from GENE runs"""

    def __init__(self, file=None, run_data=None, varname=None, prepath=None, spec=None,
                 nfields=None, extension=None):
        super().__init__(file=file, run_data=run_data, varname=varname, prepath=prepath, spec=spec,
                         extension=extension)
        self.timearray = []
        self.step = None
        self.time = None
        self.nfields = nfields
        self.fid = None
        self.extension = run_data.pnt.extension
        self.filename = self.get_filename(run_data.pnt.extension if run_data else None)
        for idx in self.varname:
            setattr(self, varname[idx] + "3d", None)

        self.loaded_tind = [-1]*len(self.varname)
        self.loaded_time = [-1]*len(self.varname)

        for idx in varname:
            new_method = File.__add_method__(self.varname[idx], idx)
            setattr(File, varname[idx], new_method)

    def _redirect(self, extension=None):
        """ Call this routine to read from a new file"""
        if extension:
            self.extension = extension
        self.filename = self.get_filename(self.extension)
        try:
            self.h5file.close()
        except (AttributeError, OSError):
            pass
        self.fid = h5py.File(self.filename, 'r')
        self.timearray = []
        self._reset_tinds()

    def get_timearray(self, extension=None):
        """ Fetch time stamps from the HDF5 file"""
        # am I pointing to the right file?
        if extension and self.extension != extension:
            self._redirect(extension)
            self.extension = extension
        if not self.fid:
            self._redirect()
        # Get time array for field file
        self.timearray = self.fid.get(self.prepath + "time")[()].tolist()
        return self.timearray

    def _readvar(self, var):
        """ Return 3d field data at the time set in self.time"""
        out = self.fid.get(
            self.prepath + self.varname[var] + "/" + '{:010d}'.format(self.tind))[()]
        # TODO convert in a different way?
        if len(out.dtype) == 2:
            out = out['real'] + 1.0j*out['imaginary']
        if len(out.shape) == 3:
            out = np.swapaxes(out, 0, 2)
        elif len(out.shape) == 4:
            np.swapaxes(out, 1, 3)
        else:
            np.swapaxes(out, 0, 1)
        if self.run_data.pnt.x_local and not self.run_data.pnt.y_local:
            # y-global has yx order
            out = np.swapaxes(out, 0, 1)
        else:
            pass
        return out


# pylint: disable=invalid-name


def GENEfile(folder, extension, file_type, parameters, spec=None):
    if parameters.pnt.write_h5:
        raise NotImplementedError('ADIOS not yet implemented')
    elif parameters.pnt.write_adios:
        raise NotImplementedError('ADIOS not yet implemented')
    else:
        try:        
            afile = BinaryFile(folder, extension, file_type, parameters, spec=spec)
            for idx, name in afile.varnames.items():
                read_method(afile, name, idx)            
            return afile
        except KeyError as kerr:
            raise ValueError('Bad file type {}'.format(file_type)) from kerr


#!/usr/bin/env python3
from tkinter import END
import os
import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as interp
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from ..putils.fourier import kx_to_x, ky_to_y
#from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagPlanes(Diagnostic):
    # pylint: disable=invalid-name
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Poloidal_planes'
        self.tabs = ['fluxtube', 'xglobal']

        self.help_txt = """Plot contours on poloidal planes
                        \n
                        \nQuantity : quantity to plot
                        \nspec: which species to plot (def. all if spec dep)
                        \nz_res: resolution in poloidal direction (def. 400)
                        \nt avg : time averaged (def. False)"""

        self.avail_vars = avail_vars

        self.specnames = specnames

        self.opts = {"quant": {'tag': "Quant", 'values': None, 'files': ['field','mom']},
                     "spec": {'tag': "Spec", 'values': self.list_specnames()},
                     "torangle": {'tag': "position", 'values': [0]},
                     "z_res": {'tag': "z res", 'values': [400]},
                     "del_zonal": {'tag': "del zonal", 'values': [False, True]},
                     "only_zonal": {'tag': "only zonal", 'values': [False, True]},
                    "t_avg": {'tag': "t avg", 'values': [False, True]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, avail_vars):
        self.avail_vars = avail_vars
        self.specnames = run_data.specnames
        if self.opts['quant']['value'] is None:
            raise RuntimeError("No quantities given for poloidal planes")
        self.torangle = self.opts['torangle']['value'] % 2*np.pi
        # Extend the z grid to pi
        self.z_full = np.append(run_data.spatialgrid.z, np.pi)
        self.z_full_fine = np.linspace(self.z_full[0], self.z_full[-1], self.opts['z_res']['value'])
        # TODO find how not to recompute is is another time
        self.triangulation = self._calc_triangulation(run_data)
        self.ny = 2*run_data.pnt.nky0
        self.dy = run_data.pnt.rhostar * run_data.pnt.minor_r*run_data.pnt.ly / self.ny

        if (run_data.pnt.x_local):
            self.q_prof = run_data.pnt.q0*(
                1 + run_data.pnt.shat/run_data.pnt.x0*(run_data.spatialgrid.x_a
                                                       - run_data.pnt.x0))
        else:
            self.q_prof = run_data.geometry.q
            
        self.Cy = run_data.geometry.Cy
        self.n0_global = run_data.pnt.n0_global
        self.get_needed_vars()
        self.get_spec_from_opts()
        return self.needed_vars

    def execute(self, data, step):
        for file in self.needed_vars:
            # loop over quaitites in that file
            for quant in self.needed_vars[file]:
                # loop over quaitites in that file
                if self.needed_vars[file][quant]:
                    if file == 'field':
                        # no species dependency
                        
                        data_in = getattr(getattr(data, file), quant)(step.time,
                                                                      getattr(step, file))
                        self.data_out = self._prepare_plane(data_in)

                    else:
                        # spec dependent
                        for spec in self.specnames:
                            data_in = getattr(getattr(data, file + '_' + spec), quant)(step.time,
                                                                                       getattr(step,
                                                                                               file))
                            self.data_out = self._prepare_plane(data_in)

    def save(self):
        print("nothing")



    def plot(self, time_requested=None, output=None, out_folder='', terminal=None):
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.set_aspect('equal', 'box')
        ax.tricontourf(self.triangulation, self.data_out.flatten(), 50)
        ax.set_xlabel(r"$R[m]$")
        ax.set_ylabel(r"$Z[m]$")
        fig.tight_layout()
        fig.show()
        if output:
            outpath = os.path.join(out_folder, output)
            fig.savefig(outpath + '.jpeg')
        print("done")

    def _calc_triangulation(self, run, recalculate=False):
        """ Calculate the triangulation based on local/global

        :param recalculate: Force recalculation of the triangulation, otherwise
        a previous result is used
        """
        if not run.pnt.y_local:
            raise RuntimeError("y-global is not supported")
        if (not recalculate):
            if run.pnt.x_local:
                return self._calc_triangulation_local(run)
            else:
                return self._calc_triangulation_xglobal(run)
        else:
            return self.triangulation

    def _calc_triangulation_xglobal(self, run):
        """ Calculate a triangulation based on R(x,z), Z(x,z) recovered from
        global geometry"""
        # These are the values at x=x0 extended so that z covers [-pi, pi]

        ## JOHNS COMMENT:  I swapped the indicies on the pad function from ((0,0),(0,1)) to ((0,1), (0,0)) and added
        ## the transpose so that dimensions matched for RectBivariateSpline constructor.  Not sure of correctness
        Rz = np.pad(run.geometry.R, ((0, 1), (0, 0)), mode="wrap").transpose()
        Zz = np.pad(run.geometry.Z, ((0, 1), (0, 0)), mode="wrap").transpose()
       

        R_pos = interp.RectBivariateSpline(run.spatialgrid.x_a, self.z_full, Rz)(run.spatialgrid.x_a, self.z_full_fine)
        Z_pos = interp.RectBivariateSpline(run.spatialgrid.x_a, self.z_full, Zz)(run.spatialgrid.x_a, self.z_full_fine)
        # Calculate a triangulation based on R(x,z) and Z(x,z)
        RZtri = mtri.Triangulation(R_pos.flatten(), Z_pos.flatten())
        # Mask central hole
        centretri = mtri.Triangulation(R_pos[0, :], Z_pos[0, :])
        centermask = np.full(RZtri.triangles.shape[0], False)
        for ct in centretri.triangles:
            start = np.where(ct.mean() == RZtri.triangles.mean(axis=1))
            centermask[start] = True
        RZtri.set_mask(centermask)
        self.triangulation = RZtri
        return self.triangulation

    def _calc_triangulation_local(self, run):
        """ Calculate a triangulation based on R(x,z), Z(x,z) recovered from
        flux-tube geometry"""
        # These are the values at x=x0 extended so that z covers [-pi, pi]
        Rz = np.pad(self.geom.R, (0, 1), mode="wrap")
        Zz = np.pad(self.geom.Z, (0, 1), mode="wrap")
        # Interpolate for the x direction
        dxdR = np.pad(self.geom.dxdR/self.geom.gxx, (0, 1), mode="wrap")
        dxdZ = np.pad(self.geom.dxdZ/self.geom.gxx, (0, 1), mode="wrap")
        x_phys = (run.spatialgrid.x_a - run.pnt.x0)*run.pnt.Lref
        R_pos = interp.RectBivariateSpline(x_phys, self.z_full, Rz + np.outer(x_phys, dxdR))(x_phys,
                                                                                             self.z_full_fine)
        Z_pos = interp.RectBivariateSpline(x_phys, self.z_full, Zz + np.outer(x_phys, dxdZ))(x_phys,
                                                                                             self.z_full_fine)
        # Calculate a triangulation based on R(x,z) and Z(x,z)
        RZtri = mtri.Triangulation(R_pos.flatten(), Z_pos.flatten())
        # Mask central hole
        centretri = mtri.Triangulation(R_pos[0, :], Z_pos[0, :])
        centermask = np.full(RZtri.triangles.shape[0], False)
        for ct in centretri.triangles:
            start = np.where(ct.mean() == RZtri.triangles.mean(axis=1))
            centermask[start] = True
        RZtri.set_mask(centermask)
        self.triangulation = RZtri
        return self.triangulation

    def _prepare_plane(self, data):
        # remove zonal or nonzonal as needed
        data = self._manage_zonal(data)
        nx, nky, nz = data.shape
        data_ext_re = self.applyBC(data)
        data_out = np.empty((nx, len(self.z_full_fine)), dtype=float)
        for i_x in range(nx):
            y_in = self.Cy[i_x]*(self.q_prof[i_x]*self.z_full_fine -
                                 self.torangle)
            # now reduce y to get the correct index
            ny = y_in / self.dy # +numel(y_in)/2
            # linear interpolation hardcoded
            cl0 = ((ny % self.ny) + self.ny) % self.ny
            d_l = cl0 - np.fix(cl0)
            d_u = 1 - d_l
            cl = np.fix(cl0)
            i_u = (((cl+1 % self.ny) + self.ny) % self.ny).astype(int)
            i_l = (((cl % self.ny) + self.ny) % self.ny).astype(int)
            posyrange = np.arange(0, self.ny)
            data_interp = interp.RectBivariateSpline(posyrange, self.z_full, data_ext_re[i_x, ...])
            data_interp_2d=  data_interp(posyrange, self.z_full_fine)
            for i_z in range(len(self.z_full_fine)):
                data_out[i_x, i_z] = (d_u[i_z]*(data_interp_2d[i_l[i_z], i_z]) +
                        d_l[i_z] * (data_interp_2d[i_u[i_z],i_z]))
        return data_out


    def applyBC(self, data_in):
        # Apply // boundary condition
        nx, nky, nz = data_in.shape
        if nky == 1:
            data_ext = np.zeros([nx, 51, nz+1],dtype=np.cdouble)
            data_ext[:, 1, 0:-1] = data_in
        else:
            data_ext = np.zeros([nx, nky, nz+1],dtype=np.cdouble)
            data_ext[:, :, 0:-1] = data_in


        for i_y in np.arange(0, nky):
                # z=pi point
                #print("DEBUG:, ", self.q_prof)
        
                data_ext[:, i_y, -1] =  data_in[:, i_y, 0] * np.exp(-2.0j * np.pi *
                                self.n0_global * i_y * self.q_prof)

                #data_ext[:, i_y, -1] =  data_in[:, i_y, 0] * np.exp(-2.0j * np.pi *
                #                self.n0_global * i_y * self.q_prof)
        return ky_to_y(data_ext, nky)

    def _manage_zonal(self, data):
        if self.opts['del_zonal']['value']:
            data[:, 0, :] = 0.0
        if self.opts['only_zonal']['value']:
            data[:, 1:-1, :] = 0.0
        return data



# -*- coding: utf-8 -*-
"""

@author: dykua, venkitesh ayyar

For downloading files from mgk_fusion in shell
"""

import sys
import os
from pymongo import MongoClient
from mgkdb.support.mgk_login import mgk_login


### Main 
def main():
    O2 = input("Please enter the server location, port, database name, username, password in order and separated by comma.\n").split(',')
    login = mgk_login(server= O2[0], port= O2[1], dbname=O2[2], user=O2[3], pwd = O2[4])
    O2_1 = input("Plese enter a target path to save the file in .pkl format. For example : ../db_credentials.pkl  \n")
    login.save(os.path.abspath(O2_1))

    database = login.connect()

    print("Successfully connected to database")
    print("Following collections exist in database",database.list_collection_names())
    print("For upload and download, you can use the credential file with the -A argument as: -A <path_to_credentials.pkl>")


if __name__=="__main__":
    main()




# -*- coding: utf-8 -*-
import struct
from os.path import getsize, join
import numpy as np

class fieldfile():
    #class constructor
    def __init__(self,file,pars):
        self.pars=pars
        self.file=file
        self.set_gridcounts()
        self.set_sizes()
        self.set_datatypes()
        self.te,self.tesize=self.TimeEntry()
        self.define_arrays()
        self.redirect(self.file)

    #call this routine to read from a new field file
    def redirect(self,file):
        self.file=file
        try:
#            close(self.f)
            self.f.close()
        except:
            pass
        self.f=open(file,'rb')
        self.tfld=[]
        self.get_timearray()
        self.reset_tinds()

    #set resolution
    def set_gridcounts(self):
        self.nx=int(self.pars['nx0'])
        self.ny=int(self.pars['nky0'])
        self.nz=int(self.pars['nz0'])

    #entry sizes
    def set_sizes(self):
        self.nfields=int(self.pars['n_fields'])
        self.intsize=4
        realsize=4+(self.pars['PRECISION']=='DOUBLE')*4
        complexsize=2*realsize
        self.entrysize=self.nx*self.ny*self.nz*complexsize
        #jumps in bytes in field/mom files
        self.leapfld=self.nfields*(self.entrysize+2*self.intsize)

    #real and complex datatypes according to endianness
    def set_datatypes(self):
        try: self.bigendian=self.pars['ENDIANNESS']=='BIG'
        except:
            self.bigendian=False
        if self.bigendian:
            self.nprt=(np.dtype(np.float64)).newbyteorder()
            self.npct=(np.dtype(np.complex128)).newbyteorder()
        else:
            self.nprt=np.dtype(np.float64)
            self.npct=np.dtype(np.complex128)

    def define_arrays(self):
        self.phi3d=np.empty((self.nz,self.ny,self.nx),dtype=self.npct)
        self.apar3d=np.empty((self.nz,self.ny,self.nx),dtype=self.npct)
        self.bpar3d=np.empty((self.nz,self.ny,self.nx),dtype=self.npct)

    def get_timearray(self):
        #get time arrays for field file
        for i in range(int(getsize(self.file)/(self.leapfld+self.tesize))):
            self.tfld.append(float(self.te.unpack(self.f.read(self.tesize))[1]))
            self.f.seek(self.leapfld,1)

    def get_minmaxtime(self):
        if not self.tfld:
            self.get_timearray()
        return self.tfld[0],self.tfld[-1]

    #defines the struct for a time entry in field
    def TimeEntry(self):
            if self.bigendian:
                    timeentry=struct.Struct('>idi')
            else:
                    timeentry=struct.Struct('=idi')
            return timeentry,timeentry.size

    #calculate offset in field file for a given timestep and variable
    def offset(self,var):
            if var in [i for i in range(self.nfields)]:
                    return self.tesize+self.tind*(self.tesize+self.leapfld)\
                +var*(self.entrysize+2*self.intsize)+self.intsize

    #returns field for given timestep
    def readvar(self,var):
        self.f.seek(self.offset(var))
        var3d=np.fromfile(self.f,count=self.nx*self.ny*self.nz,dtype=self.npct).reshape(self.nz,self.ny,self.nx)
        return var3d

    #return time index for given time, if it is present
    #otherwise, an exception is raised
    def get_tind(self):
        return self.tfld.index(self.time)

    #set current timestep
    def set_time(self,time):
        self.time=time
        self.tind=self.get_tind()

    #reset time indices
    def reset_tinds(self):
        self.tind=0
        self.ftind=[-1]*self.nfields

    #return phi. this will only read from file if necessary
    def phi(self):
        if self.ftind[0]==self.tind:
            pass
        else:
            self.ftind[0]=self.tind
            self.phi3d=self.readvar(0)
        return self.phi3d
    
    #same for apar
    def apar(self):
        if self.nfields>1:
            if self.ftind[1]==self.tind:
                pass
            else:
                self.ftind[1]=self.tind
                self.apar3d=self.readvar(1)
            return self.apar3d
    
    #same for bpar
    def bpar(self):
        if self.nfields>2:
            if self.ftind[2]==self.tind:
                pass
            else:
                self.ftind[2]=self.tind
                self.bpar3d=self.readvar(2)
            return self.bpar3d


# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import h5py
import ..utils.averages as averages
from .diagnostic import Diagnostic
from .baseplot import Plotting


class DiagDerivprofilegene3d(Diagnostic):

    # pylint: disable=invalid-name
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Deriv. Potential'
        self.tabs = ['GENE3D']

        self.help_txt = """Displays the 0th, 1st and 2nd radial derivative of the flux surface
                        \naverage of the potential vs x and t
                        \n
                        \nSave h5 : save hdf5 file (def.True) """

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"save_h5": {'tag': 'Save h5', 'values': [True, False]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, species):
        class Derivphiprof:
            def __init__(self):
                self.Phi = []
                self.Phidx = []
                self.Phi2dx2 = []

        self.geom = run_data.geometry
        self.run_data = run_data
        self.specnames = self.run_data.specnames

        self.derivphiprof_step = Derivphiprof()

        self.get_needed_vars(['field'])

        return self.needed_vars

    def execute(self, data, step):
        phi_yz_av = averages.yz_av3d(getattr(getattr(data, 'field'), "phi")(step.time, step.field),
                                     self.geom)

        x_ind = self.run_data.spatialgrid.x_a/self.run_data.pnt.rhostar

        phi_yz_av_dev = -np.gradient(phi_yz_av, x_ind, axis=0)
        phi_yz_av_2dev = np.gradient(phi_yz_av_dev, x_ind, axis=0)

        self.derivphiprof_step.Phi.append(phi_yz_av)
        self.derivphiprof_step.Phidx.append(phi_yz_av_dev)
        self.derivphiprof_step.Phi2dx2.append(phi_yz_av_2dev)

    def plot(self, time_requested, output=None, out_folder=None):
        def plot_a_map(ax, x, y, f, x_lbl, y_lbl, ttl):
            cm1 = ax.pcolormesh(x, y, f)
            ax.set_rasterization_zorder(z=-10)
            ax.set_xlabel(x_lbl)
            ax.set_ylabel(y_lbl)
            ax.set_title(ttl)
            fig.colorbar(cm1)

        self.plotbase = Plotting()

        y_lbl = r'x/a'
        x_lbl = r'$t\; [c_{ref}/L_{ref}]$'

        fig = plt.figure(figsize=(15, 10), dpi=100)

        ax = fig.add_subplot(1, 3, 1)
        ttl = r"$<\phi>_{FS}$"
        plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a,
                   np.squeeze(np.array(self.derivphiprof_step.Phi).T), x_lbl, y_lbl, ttl)

        ax = fig.add_subplot(1, 3, 2)
        ttl = r"$-\partial_{x} <\phi>_{FS}$"
        plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a,
                   np.squeeze(np.array(self.derivphiprof_step.Phidx).T), x_lbl, y_lbl, ttl)

        ax = fig.add_subplot(1, 3, 3)
        ttl = r"$-\partial^2_x <\phi>_{FS}$"
        plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a,
                   np.squeeze(np.array(self.derivphiprof_step.Phi2dx2).T), x_lbl, y_lbl, ttl)

        fig.tight_layout()
        fig.show()

        if self.opts['save_h5']['value']:
            file = h5py.File(out_folder / 'deriv_potential.h5', 'w')
            file["/x_a"] = self.run_data.spatialgrid.x_a
            file["/time"] = time_requested
            file["/phi_fs"] = np.array(self.derivphiprof_step.Phi).T
            file["/dphidx_fs"] = np.array(self.derivphiprof_step.Phidx).T
            file["/dphidx2_fs"] = np.array(self.derivphiprof_step.Phi2dx2).T
            file.close()



# -*- coding: utf-8 -*-

from tkinter import END

import h5py
import matplotlib.pyplot as plt
import numpy as np

import ..utils.averages as averages
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagFluxesgene3d(Diagnostic):

    # pylint: disable=invalid-name
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Flux profiles'
        self.tabs = ['GENE3D']

        self.help_txt = """Displays radial fluxes vs. x (averaged over time) as well as vs. x
                        \n
                        \nspec: which species to plot (def. all if spec dependent)
                        \nSave h5 : save hdf5 file (def. True)"""

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"spec": {'tag': "Spec", 'values': self.list_specnames()},
                     'save_h5': {'tag': 'Save h5', 'values': [True, False]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, species):

        class FluxStep:
            def __init__(self, specnames, electromagnetic):
                self.__em__ = electromagnetic
                for spec in specnames:
                    setattr(self, spec, self.__SpecSpectra(self.__em__))

            class __SpecSpectra:
                def __init__(self, em):
                    self.Qes = []
                    self.Ges = []

                    if em:
                        self.Qem = []
                        self.Gem = []

        self.geom = run_data.geometry
        self.run_data = run_data
        self.specnames = self.run_data.specnames

        self.flux_step = FluxStep(self.specnames, self.run_data.electromagnetic)

        self.get_needed_vars(['mom'])

        self.get_spec_from_opts()

        return self.needed_vars

    def execute(self, data, step):

        for spec in self.specnames:

            flux_step = getattr(self.flux_step, spec)

            Qes_yz_av = np.average(getattr(getattr(data, 'mom_' + spec), "Q_es")(step.time, step.mom), weights=self.geom.jacobian.T, axis=(1, 2))
            Ges_yz_av = np.average(getattr(getattr(data, 'mom_' + spec), "Gamma_es")(step.time, step.mom), weights=self.geom.jacobian.T, axis=(1, 2))

            flux_step.Qes.append(Qes_yz_av)
            flux_step.Ges.append(Ges_yz_av)

            if self.run_data.electromagnetic:

                Qem_yz_av = np.average(getattr(getattr(data, 'mom_' + spec), "Q_em")(step.time, step.mom), weights=self.geom.jacobian.T, axis=(1, 2))
                Gem_yz_av = np.average(getattr(getattr(data, 'mom_' + spec), "Gamma_em")(step.time, step.mom), weights=self.geom.jacobian.T, axis=(1, 2))

                flux_step.Qem.append(Qem_yz_av)
                flux_step.Gem.append(Gem_yz_av)

    def plot(self, time_requested, output=None, out_folder=None):

        if output:
            output.info_txt.insert(END, "Radial fluxes:\n")

        def plot_a_map(ax, x, y, f, x_lbl, y_lbl, ttl):
            cm1 = ax.pcolormesh(x, y, f)
            ax.set_rasterization_zorder(z=-10)
            ax.set_xlabel(x_lbl)
            ax.set_ylabel(y_lbl)
            ax.set_title(ttl)
            fig.colorbar(cm1)

        def plot_a_quant(ax, x, f, x_lbl, ttl):
            ax.plot(x, f.T)
            ax.set_rasterization_zorder(z=-10)
            ax.set_xlabel(x_lbl)
            ax.set_title(ttl)

        self.plotbase = Plotting()

        self.norm_q = self.plotbase.gyrobohm_SI(self.run_data, quantity="Qturb")*\
                      self.run_data.pars["Lref"] ** 2*1E-06
        self.area = averages.get_area(self.geom)

        x_lbl = r'x/a'
        t_lbl = r'$t\; [c_{ref}/L_{ref}]$'

        for spec in self.specnames:
            spec_flux = getattr(self.flux_step, spec)

            fig = plt.figure(figsize=(15, 10), dpi=100)
            ax_1 = fig.add_subplot(1, 2, 1)
            ttl = r"$" + spec + r"\,\,Q_{es}$"
            plot_a_map(ax_1, time_requested, self.run_data.spatialgrid.x_a,
                       np.squeeze(np.array(getattr(spec_flux, 'Qes')).T), t_lbl, x_lbl, ttl)
            ax_2 = fig.add_subplot(1, 2, 2)
            ttl = r"$" + spec + r"\,\,\Gamma_{es}$"
            plot_a_map(ax_2, time_requested, self.run_data.spatialgrid.x_a,
                       np.squeeze(np.array(getattr(spec_flux, 'Ges')).T), t_lbl, x_lbl, ttl)
            fig.show()

            fig = plt.figure(figsize=(15, 10), dpi=100)
            ax_1 = fig.add_subplot(2, 2, 1)
            tmp = averages.mytrapz(getattr(spec_flux, "Qes"), time_requested)
            tt1 = r"$<Q_{es}>_{FS,t}$"
            plot_a_quant(ax_1, self.run_data.spatialgrid.x_a, tmp, x_lbl, tt1)

            ax_2 = fig.add_subplot(2, 2, 2)
            tmp = averages.mytrapz(getattr(spec_flux, "Ges"), time_requested)
            tt1 = r"$<\Gamma_{es}>_{FS,t}$"
            plot_a_quant(ax_2, self.run_data.spatialgrid.x_a, tmp, x_lbl, tt1)

            ax_3 = fig.add_subplot(2, 2, 3)
            tmp = averages.mytrapz(getattr(spec_flux, "Qes"), time_requested)
            tt1 = r"$<Q_{es}>_{FS,t} * dVdx$"
            plot_a_quant(ax_3, self.run_data.spatialgrid.x_a, tmp*self.area, x_lbl, tt1)

            ax_4 = fig.add_subplot(2, 2, 4)
            tmp = averages.mytrapz(getattr(spec_flux, "Ges"), time_requested)
            tt1 = r"$<\Gamma_{es}>_{FS,t} * dVdx$"
            plot_a_quant(ax_4, self.run_data.spatialgrid.x_a, tmp*self.area, x_lbl, tt1)

            plt.tight_layout()
            fig.show()

            if self.opts['save_h5']['value']:
                file = h5py.File(out_folder / 'flux_profile_{}.h5'.format(spec), 'w')
                file["/x_a"] = self.run_data.spatialgrid.x_a
                file["/time"] = time_requested
                file["/Q_es_" + spec] = np.array(getattr(spec_flux, 'Qes')).T
                file["/G_es_" + spec] = np.array(getattr(spec_flux, 'Ges')).T
                if self.run_data.electromagnetic:
                    file["/Q_em_" + spec] = np.array(getattr(spec_flux, 'Qem')).T
                    file["/G_em_" + spec] = np.array(getattr(spec_flux, 'Gem')).T
                file.close()



#!/usr/bin/env python3
""" Command line tool for the python diagnostics"""

import argparse
import sys
import os
import matplotlib.pyplot as plt
from multiprocessing.dummy import Pool

import .pydiag.data.datafiles as datafiles
import .pydiag.data.srcmom_data
import .pydiag.data.nrgdata as nrg
import .pydiag.diagplots.plot_zonal as pzf
import .pydiag.diagplots.plot_fsamom as pfsa
import .pydiag.diagplots.plot_probdensfunc as ppdf
import .pydiag.utils.averages
import .pydiag.utils.comm as comm
import .pydiag.utils.gkdb as gkdb
import .pydiag.utils.geomfft as geomfft
import .pydiag.diagplots.plot_spectra as spectra
import .pydiag.diagplots.plot_contour as contour
import .pydiag.diagplots.plot_torus as torus
import .pydiag.diagplots.plot_ball as ballooning
import .pydiag.diagplots.plot_anisotropy as anisotropy
from .pydiag.data import profile_data, fsamom_data
from .pydiag.diagplots import plot_profiles, plot_nrg
from .pydiag.diagplots import plot_srcmom
from .pydiag.data.base_file import gluetimetraces

xrange = (None, None)
yrange = (None, None)
zrange = (None, None)

# This set controls what is plotted. TODO: Get rid of this control structure
PLOTTHIS = ['plotT', 'plotQar', 'plotQ',  # 'plotch',
            # 'plotP',
            'plotj', 'plotG', 'plotturb', 'plotnu', ]
PLOTTHIS = set(PLOTTHIS)

parser = argparse.ArgumentParser()
# Necessary parameters
parser.add_argument("time", help="Time window for averaging in the form start:end. "
                                 "Use f for the first and l for the last time in the data file.")
parser.add_argument("fileextension", nargs='+', help="List of file extensions of the runs"
                                                     " to be processed.")

# Options related to file handling
parser.add_argument("--runpath", default=".", help="Path of the run")
parser.add_argument("--continuation", "-c", action='store_true',
                    help="Try to combine time traces into one object"
                         "This requires matching parameter sets")

# Options for the diagnostic grid to use
gridopts = parser.add_argument_group("Diagnostic grid options")
gridopts.add_argument("--xavg", help="Average over full x direction", action="store_true")
gridopts.add_argument("--yavg", help="Average over full y direction", action="store_true")
gridopts.add_argument("--zavg", help="Average over full z direction", action="store_true")
gridopts.add_argument("--xfourier", help="Operate in kx space", action="store_true")
gridopts.add_argument("--yfourier", help="Operate in ky space", action="store_true")
gridopts.add_argument("--zfourier", help="Operate in kz space", action="store_true")
gridopts.add_argument("--species", "-s", help="Names of species to process", nargs='+')

# Auxiliary options to request particular index values (/ranges)
indexargs = parser.add_argument_group("Index values / ranges")
indexargs.add_argument("--kyind", type=int, nargs=1, #action="store_const",
                       default=None,
                       help="ky index to be considered in diagnostics (if supported)")


# Output options
outputargs = parser.add_argument_group("Output options")
outputargs.add_argument("--save", action='store_true',
                        help="Save the plots as pdfs or videos (for animations)")
outputargs.add_argument("--timetrace", action='store_true',
                        help="Generate text files with time traces "
                             "(currently only T at 3 fixed x/a) from profiles")
outputargs.add_argument("--averageprofiles", action='store_true',
                        help="Generate a time averaged profile suitable for GENE input")
outputargs.add_argument("--gkdb", action='store_true',
                        help="Save data for GKDB (github.com/gkdb/gkdb)")

outputargs.add_argument("--geomfft", action='store_true',
                        help="Parametrize GENE flux surface by Fourier coefficients")

# Different types of diagnostics
parser.add_argument("--noturb", "-t", action='store_true',
                    help="Exclude turbulent transport in profile plots")
diagopts = parser.add_mutually_exclusive_group()
diagopts.add_argument("--profxt", "-x", action='store_true',
                      help="Create x-t contour plots of profiles")

diagopts.add_argument("--zonal", "-z", action='store_true',
                      help="Plot average and x-t contour of zonal phi"
                           "instead of anything else")
diagopts.add_argument("--nrg", "-n", action='store_true', help="Plot nrg file with average and "
                                                               "uncertainty estimate")
diagopts.add_argument("--probabilitydens", "-d", action='store_true',
                      help="Generate a plot of the probability density function for fluxes "
                           "in profiles")
diagopts.add_argument("--fsa", action='store_true', help="Plot time averaged plots of the"
                                                         " FSA moment diagnostic")
diagopts.add_argument("--srcmom", action="store_true", help="Plot time averaged source moments")
diagopts.add_argument("--spectra", action="store_true",
                      help="Plot amplitude spectra of mom and field")
diagopts.add_argument("--perpspectra", action="store_true",
                      help="Plot 1d amplitude spectra of fields vs. perpendicular wavenumber")
diagopts.add_argument("--fluxspectra", action="store_true",
                      help="Plot flux spectra based on mom and field")
diagopts.add_argument("--xycontours", action="store_true",
                      help="Contour plots in x and y based on mom and field")
diagopts.add_argument("--ballooning", "-ball", action="store_true",
                      help="Plot mode structure for all fields")
diagopts.add_argument("--anisotropy", action='store_true', help="Plot scale-dependent anisotropy "
                      "of magnetic field fluctuations")
diagopts.add_argument("--toruscut", action='store_true', help="Plot contours on"
                                                              " a cut through the torus")

args = parser.parse_args()
os.chdir(args.runpath)

fileextension = args.fileextension

if args.noturb:
    PLOTTHIS.discard('plotturb')

if args.time.split(':')[0] == 'f':
    starttime = -1
elif args.time.split(':')[0] == 'l':
    starttime = -2
else:
    starttime = float(args.time.split(':')[0])
if args.time.split(':')[1] == 'f':
    endtime = -1
elif args.time.split(':')[1] == 'l':
    endtime = -2
else:
    endtime = float(args.time.split(':')[1])

inputcomm = []

inputdata = []
inputdiagspace = []
inputfsa = []
inputfield = []
inputmom = []
inputsrcmom = []
inputnrg = []
inputprof = []

speclist = []
for fext in fileextension:
    common = comm.CommonData(fext, starttime, endtime)
    inputcomm.append(common)
    if not args.species:
        speclist += common.specnames
    else:
        speclist = args.species
    diagspace = comm.DiagSpace(common.spatialgrid, x_fourier=args.xfourier, y_fourier=args.yfourier,
                               z_fourier=args.zfourier, xrange=xrange, yrange=yrange, zrange=zrange,
                               xavg=args.xavg, yavg=args.yavg, zavg=args.zavg)
    inputdiagspace.append(diagspace)
    rundatafiles = datafiles.RunDataFiles(common)
    inputdata.append(rundatafiles)

# Remove all double entries from species list
speclist = list(set(speclist))

if args.nrg:
    for icomm, common in enumerate(inputcomm):
        filename = 'nrg{}'.format(common.fileextension)
        nrgdat = inputdata[icomm].get_fileobject("nrg")
        inputnrg.append(nrgdat)
    if args.continuation:
        inputnrg = [nrg.gluenrgdata(inputnrg)]
    plotn = plot_nrg.PlotNrgdata(inputnrg)
    plotn.print_mean_err()
    plotn.plot_ttrace(poutput=args.save)


elif args.zonal:
    for icomm, common in enumerate(inputcomm):
        fielddat = pzf.ZonalAverage(common, inputdata[icomm], xrange=xrange)
        inputfield.append(fielddat)
    if args.continuation:
        inputfield = [gluetimetraces(inputfield)]
    zon = pzf.PlotZonalAverage(inputcomm, inputfield)
    zon.createfigs(withext=True, poutput=args.save)
    plotf = pzf.PlotZonalxt(inputcomm, inputfield)
    plotf.plot_phi_xt(poutput=args.save)
elif args.fsa:
    for common in inputcomm:
        try:
            fsadat = fsamom_data.FSAmomData(common)
            fsadat.getfsadata()
        except IOError:
            fsadat = fsamom_data.FSAmomData_legacy(common)
            fsadat.getfsadata()
        inputfsa.append(fsadat)
    plotfsa = pfsa.PlotFSA(inputfsa)
    plotfsa.createfigs(poutput=args.save)
elif args.srcmom:
    for icomm, common in enumerate(inputcomm):
        srcmom_f0 = pydiag.data.srcmom_data.SrcmomSeries(common, inputdata[icomm], "energy", "f0")
        srcmom_ckh = pydiag.data.srcmom_data.SrcmomSeries(common, inputdata[icomm], "energy",
                                                          "ck_heat")
        srcmom = [srcmom_f0, srcmom_ckh]
        inputsrcmom.append(srcmom)
    inputsrcmom = [item for sublist in inputsrcmom for item in sublist]

    plotsrc = plot_srcmom.PlotSrcmomAverage(inputsrcmom)
    plotsrc.createfigs()

elif args.spectra:
    inputampspec = []
    for icomm, common in enumerate(inputcomm):
        with Pool(len(speclist)) as p:
            inputampspec = p.map(
                    lambda spec: spectra.AmplitudeSpectra(common, spec, inputdata[icomm]),
                    speclist)
    specplt = spectra.PlotAmplitudeSpectra(inputcomm, inputampspec)
    specplt.createfigs()

elif args.perpspectra:
    inputperpspec = []
    for icomm, common in enumerate(inputcomm):
        with Pool(len(args.species)) as p:
            inputperpspec = p.map(
                    lambda spec: spectra.PerpSpectra(common, spec, inputdata[icomm]),
                    args.species)
    specplt = spectra.PlotPerpSpectra(inputcomm, inputperpspec)
    specplt.createfigs()

elif args.fluxspectra:
    inputfluxspec = []
    for icomm, common in enumerate(inputcomm):
        with Pool(len(speclist)) as p:
            inputfluxspec = p.map(lambda spec: spectra.FluxSpectra(common, spec, inputdata[icomm],
                                                                   with_momentum=False),
                                  speclist)
    specplt = spectra.PlotFluxSpectra(inputcomm, inputfluxspec)
    specplt.createfigs()
    for s in inputfluxspec:
        s.print_total_fluxes()

elif args.xycontours:
    inputcontours = []
    for icomm, common in enumerate(inputcomm):
        with Pool(len(speclist)) as p:
            inputcontours = p.map(
                    lambda spec: contour.MomFieldContour(common, spec, inputdata[icomm],
                                                         zavg=args.zavg,
                                                         quantities=["phi"]), speclist)
    contplot = contour.PlotXYContour(inputcontours)
    contplot.createfigs(animate=True, save=args.save)

elif args.toruscut:
    inputtorus = []
    for icomm, common in enumerate(inputcomm):
        spec = speclist[0]
        if len(speclist) != 1:
            print("Using only first given species for torus cut: {}".format(spec))
        inputtorus.append(torus.TorusContourCut(common, spec, inputdata[icomm], quantity="phi"))
    torusplot = torus.PlotTorusCut(inputtorus)
    torusplot.createfigs(animate=True, save=args.save)

elif args.ballooning:
    for icomm, common in enumerate(inputcomm):
        fielddat = ballooning.ModeStructure(common, inputdata[icomm], kyind=args.kyind,normalize=True)
        inputfield.append(fielddat)
    if args.continuation:
        inputfield = [gluetimetraces(inputfield)]
    ballplot = ballooning.PlotModeStructure(inputcomm, inputfield)
    ballplot.createfigs()

elif args.gkdb:
    for icomm, common in enumerate(inputcomm):
        gkdb_obj = gkdb.GKDB(common, inputdata[icomm])
        inputfield.append(gkdb_obj)
    if args.continuation:
        inputfield = [gluetimetraces(inputfield)]
    gkdb_obj.Write_json('./')
elif args.geomfft:
    for icomm, common in enumerate(inputcomm):
        geomfft_obj = geomfft.GEOMFFT(common, inputdata[icomm])
        inputfield.append(geomfft_obj)
elif args.anisotropy:
    for icomm, common in enumerate(inputcomm):
        fielddat = anisotropy.Anisotropy(common, inputdata[icomm])
        inputfield.append(fielddat)
#    if args.continuation:
#        inputfield = [gluetimetraces(inputfield)]
    aniplot = anisotropy.PlotAnisotropy(inputcomm, inputfield)
    aniplot.createfigs()

else:
    for icomm, common in enumerate(inputcomm):
        prdat = profile_data.ProfileData(common, inputdata[icomm])
        fielddat = pydiag.diagplots.plot_zonal.ZonalAverage(common, inputdata[icomm], xrange=xrange)
        prdat.get_profiles()
        inputprof.append(prdat)
        if not (args.averageprofiles or args.timetrace or args.probabilitydens or args.profxt):
            fielddat.generate_timeseries()
            inputfield.append(fielddat)
    if args.continuation:
        inputprof = [profile_data.glueprofiledata(inputprof)]
        if not (args.averageprofiles or args.timetrace):
            inputfield = [gluetimetraces(inputfield)]
    if args.timetrace:
        for prdat in inputprof:  # Time trace generation
            prdat.generate_timetrace(0.45, "Qturb")
            prdat.generate_timetrace(0.6, "Qturb")
            prdat.generate_timetrace(0.75, "Qturb")

    if args.profxt:
        plot = plot_profiles.PlotProfxt(inputprof)
        plot.createfigs(args.save)

    elif args.probabilitydens:
        plotpd = ppdf.PlotProfilePdf(inputprof)
        plotpd.createfigs(poutput=args.save)

    else:
        plot = plot_profiles.PlotProfdata(inputprof, inputfield, PLOTTHIS)
        plot_profiles.printaverageflux(inputprof, 0.4, 0.6)
        if args.averageprofiles:
            plot.output_avprof("average")
            sys.exit("Only average profile generated")
        plot.createfigs(args.save)

plt.show()



""" Module containing fourier transform routines for GENE time series"""

import numpy as np


def kx_to_x(var_kx, nx0, axis=-3):
    """ Perform inverse FFT on kx spectral direction of variable

    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    :param var_kx: Variable in kx space
    :param nx0: Number of kx grid points
    :param axis: Which axis of var_kx is the x direction, by default the third last one
    :returns: variable in real x space
    """
    var_x = np.fft.ifft(nx0*var_kx, axis=axis)
    var_x = np.real_if_close(var_x, tol=1e5)
    return var_x


def ky_to_y(var_ky, nky0, axis=-2):
    """ Perform inverse FFT on ky spectral direction of variable

    The GENE data only include the non-negative ky components, so we need to use the real
    valued FFT routines
    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    :param var_ky: Variable in kx space
    :param nky0: Number of ky grid points
    :param axis: Which axis of var_kx is the x direction, by default the third last one
    :returns: variable in real y space
    """
    # The GENE data only include the non-negative ky components, so we need to use the real
    # valued FFT routines
    var_y = np.fft.irfft(2*nky0*var_ky, n=2*nky0, axis=axis)
    var_y = np.real_if_close(var_y, tol=1e5)
    return var_y


def kz_to_z(var_kz, nz0, axis=-1):
    """ Perform inverse FFT on kz spectral direction of variable

    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    This does not apply to the z direction though, since there is no z-spectral
    version of GENE.
    :param var_kz: Variable in kz space
    :param nz0: Number of kz grid points
    :param axis: Which axis of var_kz is the z direction, by default the third last one
    :returns: variable in real z space
    """
    var_z = np.fft.ifft(var_kz, axis=axis)
    var_z = np.real_if_close(var_z, tol=1e5)
    return var_z


def x_to_kx(var_x, nx0, axis=-3):
    """ Perform FFT on x direction of variable

    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    :param var_x: Variable in real x space
    :param nx0: Number of x grid points
    :param axis: Which axis of var_kx is the x direction
    :returns: variable in fourier kx space
    """
    var_kx = np.fft.fft(var_x, axis=axis)
    return var_kx


def y_to_ky(var_y, nky0, axis=-2):
    """ Perform FFT on y direction of variable

    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    :param var_y: Variable in real y space
    :param nky0: Number of y grid points
    :param axis: Which axis of var_ky is the y direction
    :returns: variable in fourier ky space
    """
    var_ky = np.fft.rfft(var_y, axis=axis)
    return var_ky


def z_to_kz(var_z, nz0, axis=-1):
    """ Perform FFT on z direction of variable

    Note: The fft and ifft in python and GENE/IDL have the factor 1/N switched!
    :param var_z: Variable in real z space
    :param nz0: Number of z grid points
    :param axis: Which axis of var_kz is the z direction
    :returns: variable in fourier kz space
    """
    var_kz = np.fft.fft(var_z, axis=axis)
    return var_kz



""" Module with the container objects defining velocitry grid and weights"""
# -*- coding: utf-8 -*-
import numpy as np
from .par_io import Parameters

class VspGrid:
    def __init__(self, parameters):
        if isinstance(parameters, Parameters):
            pnt = parameters.asnamedtuple()
        elif isinstance(parameters, tuple):
            pnt = parameters

        self._calc_vgrid(pnt.lv, pnt.nv0)

        if pnt.mu_grid_type == 'gau_lag' or pnt.mu_grid_type == "'gau_lag'":
            self._calc_mugrid_gaulag(pnt.lw, pnt.nw0)
        elif pnt.mu_grid_type == "'eq_vperp'":
            self._calc_mugrid_eq_vperp(pnt.lw, pnt.nw0)
        else:
            print("mu grid " + pnt.mu_grid_type + " is not yet implemented")

    def _calc_vgrid(self, lv, nv):
        dv = 2.0*lv/(nv-1.0)
        self.vp = np.zeros(nv)
        for iv in range(0, nv):
            self.vp[iv] = -lv + iv*dv

    def _calc_mugrid_gaulag(self, lw, nw):
        self.mu, dmu = np.polynomial.laguerre.laggauss(nw)
        dmu = dmu * np.exp(self.mu)
        fac = lw / np.sum(dmu)
        self.mu = self.mu*fac

    def _calc_mugrid_eq_vperp(self, lw, nw):
        """Sugama needs this"""
        dmu = lw / nw**2
        self.mu = np.zeros(nw)
        self.muweights = np.zeros(nw)
        for iw in range(1, nw+1):
            self.mu[iw-1] = (iw + 0.5)**2 * dmu
            self.muweights[iw-1] = (2.*iw - 1.) * dmu



# Code to download all non-linear runs in database 

import sys
import os
import argparse
from sys import exit

import gridfs
import json 
from bson.objectid import ObjectId

from mgkdb.support.mgk_login import mgk_login,f_login_dbase

# Run this as : 
# python update_meta_code.py -A <fname.pkl> -C linear -m 3

def f_parse_args():
    parser = argparse.ArgumentParser(description='Update Metadata entries. 3 modes.  1: Append to publication list.\n2: Append to comments.\n 3: Update any specific entry. \n Modes 1 and 2 add entered values to existing entry.\nUse mode=3 with caution as you are rewriting previous entry.')

    parser.add_argument('-C', '--collection', choices=['linear','nonlinear'], default='linear', type=str, help='collection name in the database')
    parser.add_argument('-OID', '--objectID', default = '678a700bd978ec23d2ebfd18', help = 'Object ID in the database')
    parser.add_argument('-m', '--mode', type=int, choices=[1,2,3], default = 3, help = 'Choose mode of operation for updating Metadata. 1: Append to publication list.\n2: Append to comments.\n 3: Update any specific entry.')
    parser.add_argument('-A', '--authenticate', default = None, help='locally saved login info, a .pkl file')

    return parser.parse_args()

def f_update_metadata(data, key_name, template_d):
    '''
    Update the existing metadata key entry 
    key_name : key to update 
    template_d : dictionary containing template
    '''
    if isinstance(data, dict):
        new_dict ={}
        for key,value in data.items():
            new_val = f_update_metadata(value, key, template_d)
            new_dict[key] = new_val

        return new_dict 
    
    elif isinstance(data, list):
        ## Create dictionary using last value 
        new_val = f_update_metadata(data[-1],key_name, template_d)
        
        return new_val

    else:
        print("The existing entry for this key is:\t%s"%(data))
        ## if entry is string, just replace
        new_value = input('Please enter the new entry you want for %s. If you don\'t want to proceed, enter : none \n'%(key_name))

        return new_value

if __name__=="__main__":
    
    ### Parse arguments 
    args = f_parse_args()
    Args = vars(args)
    
    authenticate = Args['authenticate']
    collection_ip = Args['collection']
    
    ### Connect to database 
    login = f_login_dbase(authenticate)
    database = login.connect()

    #### Dict to convert from argument to collection name in database
    collection_dict={'linear':'LinearRuns','nonlinear':'NonlinRuns','files':'fs.files'}
    collection =  getattr(database,collection_dict[collection_ip])
    
    # Get Object IDs 
    all_ids = [r['_id'] for r in collection.find({},{'id':1})]
    
    oid = ObjectId(args.objectID)
    assert oid in all_ids,"Given oid entry %s doesn't exist in database"%(oid)
    
    ### Extract document for this oid
    document = collection.find_one({"_id":oid},{'Metadata':1,'_id':0})

    ### Check user credential
    uname = login.login['user']
    input_usr = document['Metadata']['DBtag']['user']

    if uname!=input_usr:
        cont = input(f"Data was input by another use {input_usr}. Do you wish to continue ? y or n \n")
        if not cont=='y': 
            raise SystemExit
        
    keys = document.get('Metadata').keys()

    ### Options : Append to publications list, Append to comments, Update any specific entry
    if args.mode==1: 
        user_ip = input('Enter the doi strings to append, separated by commas\n')
        entry_add = user_ip.split(',')

        ## Existing entry
        entry_val = document['Metadata']['Publications']['doi']
        if not entry_val: entry_val = [] 
        
        new_pub_list = entry_val + entry_add ## Join the two lists 

        assert isinstance(new_pub_list,list),f"Entry to add must be a list"

        fltr = {"_id": oid}
        update = {"$set": {"Metadata.Publications.doi": new_pub_list}}
        result = collection.update_one(fltr, update)
        print("Appended publication record")

    elif args.mode==2: # Option to append to comment string
        old_comment = document['Metadata']['DBtag']['comments']
        # assert isinstance(old_comment,str),f"Existing entry {old_comment} is not a string" 
        user_ip = input(f'Enter the string to append to current entry. Current entry is: \n {old_comment} \n')
        
        fltr = {"_id": oid}
        # Using $set to add appended string to comments
        update = {"$set": {"Metadata.DBtag.comments": old_comment+'.\n'+user_ip}}
        result = collection.update_one(fltr, update)
        print("Appended comment")
        
    elif args.mode==3: 
        top_key = input(f'Please enter the key in Metadata that you want to update. Allowed options are: \n {list(keys)}\n')
        assert top_key in keys,f"Invalid input key {top_key}"

        sub_keys = document.get('Metadata').get(top_key).keys()
        key_name = input(f'The key {top_key} has the following subkeys {list(sub_keys)}. \nPlease select which subkey you want to modify\n')
        assert key_name in sub_keys,f"Invalid input key {key_name}"

        ans = document.get('Metadata').get(top_key).get(key_name)
        print("The existing entry for this key is:\t%s"%(ans))
        print("You will be resetting the entire value for this key. Please use caution")          
        new_value = input('Please enter the new entry you want for %s. If you don\'t want to proceed, enter : none \n'%(key_name))
        confirm = input(f'Confirm changing entry to: \n{new_value}? \n Enter Y or N\n').strip().upper()

        ## Some entries need to be in a list 
        if top_key in ['Publications']:   new_value = [new_value]

        if (confirm!='Y' or (new_value == 'none')):
            print('Didn not receive confirmation. Aborting update')
            raise SystemExit
            
        fltr = {"_id": oid}
        update = {"$set": {f"Metadata.{top_key}.{key_name}": new_value}}
        result = collection.update_one(fltr, update)

        document = collection.find_one({"_id":oid},{'Metadata':1,'_id':0})
        print('Updated entry: %s '%(document.get('Metadata').get(top_key).get(key_name)))    





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  6 16:29:16 2019

"""
import numpy as np


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def find_nearest_points(array, val_min, val_max):

    array = np.asarray(array)
    idx_min = (np.abs(array - val_min)).argmin()
    idx_max = (np.abs(array - val_max)).argmin()

    if idx_max == 0:
        idx_min = 0
        idx_max = 1

    if idx_min >= idx_max:
        if (idx_min + 1) == len(array):
            idx_min = idx_max - 1
        else:
            idx_max = idx_min + 1

    return idx_min, idx_max



# -*- coding: utf-8 -*-

import warnings
from tkinter import END
import matplotlib.pyplot as plt
import numpy as np
from ..putils import fourier
from .diagnostic import Diagnostic
from .baseplot import Plotting
from .diagspace import DiagSpace
import os

class DiagCrossPhase(Diagnostic):
    """Diagnostic for the cross phases between pairs of variables in ky space."""
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Cross phase'
        self.tabs = ['fluxtube']

        self.help_txt = """Plots the PDFs of the cross phases between pairs of variables in ky space.
                        \nQuantity 1: First quantity for the cross phase
                        \nQuantity 2: Second quantity for the cross phase
                        \nxind: x indices for cross phase calculation; default: -1 (average)
                        \nzind: z point for cross phase calculation; default: -1 (average)
                        \nspec: which species to plot (def. all if spec dependent)
                        \nSave h5 : save hdf5 file (def. False)
                        """

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"quant_1": {'tag': "Quant 1", 'values': None, 'files': ['field']},
                     "quant_2": {'tag': "Quant 2", 'values': None, 'files': ['mom']},
                     "spec": {'tag': "Spec", 'values': self.list_specnames()},
                     "xind": {'tag': "x ind", 'values': None},
                     "zind": {'tag': "z ind", 'values': None}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run, specnames, out_folder):
        
        self.ky = run.spatialgrid[0].ky
        self.pnt = run.parameters[0].asnamedtuple()

        if self.opts['quant_1']['value'] is None:
            raise RuntimeError("No quantity 1 given for cross phase")
        if self.opts['quant_2']['value'] is None:
            raise RuntimeError("No quantity 2 given for cross phase")

        # parallel position, unset or -1, average all
        zind = self.opts['zind']['value']
        if not zind or zind == -1:
            zind = None
            zrange = (None, None)
            zavg = True
        else:
            zrange = (zind, zind + 1)
            zavg = False

        # radial position, unset or -1 means all
        xind = self.opts['xind']['value']
        if not xind or xind == -1:
            xind = None
            xrange = (None, None)
            xavg = True
        else:
            xrange = (xind, xind + 1)
            xavg = False

        # binormal position, for all
        yind = None
        yrange = (None, None)
        yavg = False

        # diagspace is unique for each call, so it can be here
        # it should however be outside, since we can construct all grids
        # and have them in common for all diagnostics and select here only
        # the indexes we weant to plot
        self.diagspace = DiagSpace(run.spatialgrid[0], False, True, False, xrange, yrange, zrange,
                                   xavg, yavg, zavg)

#        self.geom = run.geometry
#        self.run_data = run_data

        # This is special since we dont save stuff by default, so just do what we need
        self.get_spec_from_opts()
        # toggle the reader
        for file_key in self.avail_vars.keys():
            self.needed_vars[file_key] = {}
            for data_value in self.avail_vars[file_key].values():
                if data_value in self.opts['quant_1']['value']:
                    self.needed_vars[file_key][data_value] = True
                elif data_value in self.opts['quant_2']['value']:
                    self.needed_vars[file_key][data_value] = True
                else:
                    self.needed_vars[file_key][data_value] = False

        self.cross_phase = {}
        for spec in self.specnames:
            self.cross_phase[spec] = []

        return self.needed_vars

    def execute(self, data, parameters, geometry, spatialgrid,  step):

        # loop over quaitites in that file
        for quant_1 in self.needed_vars["field"].keys():
            # loop over quaitites in that file
            if self.needed_vars["field"][quant_1]:
                data_in_1 = getattr(getattr(data, "field"), quant_1)(step.time,
                                                                     getattr(step, "field"), geometry)
                data_in_1 = fourier.apply_fouriertransforms(self.pnt, self.diagspace, data_in_1, geometry)

        # loop over quaitites in that file
        for quant_2 in self.needed_vars["mom"].keys():
            # loop over quaitites in that file
            if self.needed_vars["mom"][quant_2]:
                # spec dependent
                for spec in self.specnames:
                    data_in_2 = getattr(getattr(data, "mom" + '_' + spec), quant_2)(step.time,
                                                                                    getattr(step,
                                                                                            "mom"), geometry)
                    data_in_2 = fourier.apply_fouriertransforms(self.pnt, self.diagspace, data_in_2, geometry)

                    angle = np.angle(data_in_1/data_in_2)

                    # not where when we have to do the averages
                    # cross_phase = averages.av3d_by_switch(self.diagspace.xavg,
                    #                         self.diagspace.yavg,
                    #                         self.diagspace.zavg)(angle, self.geom)

                    self.cross_phase[spec].append(angle[self.diagspace.diagslice])
    
    def dict_to_mgkdb(self):
        Diag_dict = {}
        Diag_dict['nx0'] = self.pnt.nx0
        Diag_dict['nz0'] = self.pnt.nz0
        Diag_dict['cross_phase'] = self.cross_phase
        
        return Diag_dict

    def plot(self, time_requested, output=None, out_folder=None, terminal=None, suffix = None):
        """ Cross-phases """

        if output:
            output.info_txt.insert(END, "Cross phase:\n")

#        ky = self.ky
        nbins = 64

        out = np.zeros((len(self.ky), nbins))

        self.plotbase = Plotting()

        for spec in self.specnames:
            array = np.array(self.cross_phase[spec])           
            for i_ky in range(len(self.ky)):
                phase_ky = array[:,:,i_ky,...].flatten()
#                phase_ky = self.cross_phase[spec][:][:][i_ky][:][:].flatten()
                hist, bin_edges = np.histogram(phase_ky/self.pnt.nx0/self.pnt.nz0,
                                               bins=nbins, range=(-np.pi, np.pi))
                out[i_ky, :] = hist

            fig = plt.figure()
            plt.xlabel('Phase angle/rad', fontsize=14)
            plt.ylabel(r'$k_y\rho_s$', fontsize=16)
            # cm1 = plt.contourf(bin_edges[1:],ky[1:],out[1:,:],50,cmap=self.plotbase.cmap_bidirect)
            cm1 = plt.pcolormesh(bin_edges[1:], self.ky[1:], out[1:, :])
            fig.colorbar(cm1)
            fig.tight_layout()
            if out_folder is not None:
                fig.savefig(os.path.join(out_folder, 'CrossPhase_{}{}.png'.format(spec, suffix)), bbox_inches='tight')
                plt.close(fig)
            else:
                fig.show()
            
    def save(time_requested, output=None, out_folder=None):
        pass

    def finalize():
        pass



import glob
import os
import csv
import numpy as np
import h5py
import pathlib
from putils.par_io import Parameters
from putils.plotter import *

import matplotlib.pyplot as plt
from putils.time_traces import find_nearest, mytrapz


class NrgData():
    """NrgData: Class to read a nrg file from GENE
    """
    def __init__(self, folder=None, extensions=None):       
        self.data = []
        self.time = []
        self.dsets = []
        self.params = []
        
        if folder:
            self.folder = folder
        else:
            raise Exception("Don't know where to read from, provide a path")    
            
        if extensions:
            self.extensions=extensions if type(extensions) is list else [extensions]    
        else:
            self.extensions=self.__find_runs()
            
        for ext in self.extensions:
            self.params.append(Parameters(self.folder, ext))
            self.dsets.append(self.__set_dsets(self.params[-1]))
            self.__load_single(ext, self.params[-1])
            
        
        if self.__should_concatenate():
            self.__concatenate()
        
    def __find_runs(self):
        """parse the folder for all files with extension nrg 
        Could do the same for hdf5, but nrg is always written
        """
        ext=[]
        for file in glob.glob(self.folder+'/nrg*'):
            head, tail = os.path.split(file)
            ext.append(tail.replace('nrg',''))
        return ext
                       
    def __should_concatenate(self):
        """ Decide if I should concatenate multiple files
        """        
        if isinstance(self.folder, str):
            file = pathlib.Path(self.folder+"scan.log")
        else:
            file = self.folder / ("scan.log")
        return (not file.exists()) and len(self.extensions)>1

    def __set_dsets(self, params):
        """set content for plotting and h5 read
        """
        dsets = {0: ['n2', r"$n^2$"],
                 1: ['u_par2',r'$u_{||}^2$'],
                 2: ['T_par',r'$T_{||}$'],
                 3: ['T_par',r'$T_{\perp}$'],
                 4: ['Gamma_es',r'$\Gamma_{es}$'],
                 5: ['Gamma_em',r'$\Gamma_{em}$'],
                 6: ['Q_es',r'$Q_{es}$'],
                 7: ['Q_em',r'$Q_{em}$'],
                 8: ['P_es',r'$\Pi_{es}$'], 
                 9: ['P_em',r'$\Pi_{em}$']}
                     
        return dict(list(dsets.items())[0:params.pnt.nrgcols] )
    
    def __load_single(self, extension, parameters):
        """ load a single file
        """
        if parameters.pnt.write_adios:
            raise NotImplementedError('ADIOS not yet implemented')
        if parameters.pnt.write_h5:
            fname = os.path.join(self.folder, 'nrg' + extension + '.h5')
            if not os.path.isfile(fname):
                raise Exception(os.path.abspath(fname) + " does not exist")
            dsets = self.__set_dsets(parameters.pnt.nrgcols)
            nrgfile = h5py.File(fname, 'r')
            timearray = nrgfile.get("/nrg{}".format(parameters.specs[0]) +
                                    "/time")[()]
            dataarray = np.zeros(timearray.size, parameters.pnt.n_spec,
                                 parameters.pnt.nrgcols)
            for i_d, dset in enumerate(dsets):
                for ispec, spec in enumerate(self.specnames):
                    self.data[:, ispec, i_d] = nrgfile.get("/nrg" + spec + "/" + dset)[()]
            nrgfile.close()
        else:
            #open file read as csv, build numpy arrays.
            fname = os.path.join(self.folder, 'nrg' + extension)
            if not os.path.isfile(fname):
                raise Exception(os.path.abspath(fname) + " does not exist")
            data = []
            time = []   
            with open(fname) as nrgfile:
                csvnrg = csv.reader(nrgfile, delimiter=' ', skipinitialspace=True)
                for line in csvnrg:
                    if len(line) == 0:
                        continue
                    if len(line) == 1:
                        time.append(float(line[0]))
                        data.append([[] for _ in range(parameters.pnt.n_spec)])
                        ispec = 0
                    elif len(line) == parameters.pnt.nrgcols:
                        data[-1][ispec] = line
                        ispec += 1
                    else:
                        raise IOError("Incorrect number of columns")
            self.data.append(np.array(data).astype(float, copy=False))
            self.time.append(np.array(time).astype(float, copy=False))         

    def __concatenate(self):
        """ sort and concatenate traces"""             
        self.__sort_extensions()                              
                     
        time = self.time[0]
        data = self.data[0]
        for i_ext in np.arange(1,len(self.extensions)):
            # When times overlap, throw away repeats from previous
            mask = time < self.time[i_ext][0]
            time = time[mask]
            data = data[mask, ...]
            time = np.concatenate((time, self.time[i_ext]))
            data = np.concatenate((data, self.data[i_ext]))
        self.data=[data]
        self.time=[time]
        # this will not support version jumps
        self.dsets=[self.dsets[0]]
        self.params=[self.params[0]]               
                        
    def __sort_extensions(self):
        """ sort the extensions in ascending order"""           
        t_start=np.empty([0])    
        for i_ext, ext in enumerate(self.extensions):
            t_start=np.append(t_start,self.time[i_ext][0]) if t_start.size else self.time[i_ext][0]
        ordered = np.argsort(t_start)
        self.params = [x for _, x in sorted(zip(ordered, self.params))]
        self.extensions = [x for _, x in sorted(zip(ordered, self.extensions))]
        self.dsets = [x for _, x in sorted(zip(ordered, self.dsets))]
        self.data = [x for _, x in sorted(zip(ordered, self.data))]
        self.time = [x for _, x in sorted(zip(ordered, self.time))]

    def plot_fluxes(self, window=None):
        def plot_single(time, data, params, data_lbl):
            n_rows = 2 + int(params.pnt.nrgcols == 10)
            n_cols = params.pnt.n_spec
            if window:
                figure = window
            else:
                figure = plt.figure(figsize=(14,6), dpi= 100, facecolor='w', edgecolor='k')
            for i_sp in range(0, n_cols):
                # heat flux
                ax = figure.add_subplot(n_rows, n_cols, i_sp + 1)
                ax.plot(time, data[:,i_sp,6],'-b',label=data_lbl[6][1])
                if params.pnt.beta>0:
                    ax.plot(time, data[:,i_sp,7],'-m',label=data_lbl[7][1])
                    ax.plot(time, np.zeros_like(time),color='black', linewidth=1)
                set_scales(ax,params)
                ax.autoscale(enable=True, axis='x',tight=True)
                ax.legend(); ax.set_title(params.pnt.specnames[i_sp])
                # particle flux
                ax = figure.add_subplot(n_rows, n_cols, i_sp + n_cols +1)
                ax.plot(time, data[:,i_sp,4],'-b',label=data_lbl[4][1])
                if params.pnt.beta>0:
                    ax.plot(time, data[:,i_sp,5],'-m',label=data_lbl[5][1])
                    ax.plot(time, np.zeros_like(time),color='black', linewidth=1)
                set_scales(ax,params)
                ax.autoscale(enable=True, axis='x',tight=True)
                ax.legend();
                if params.pnt.nrgcols == 10:
                    # parallel momentum flux
                    ax = figure.add_subplot(n_rows, n_cols, i_sp + 2*n_cols + 1)
                    ax.plot(time, data[:,i_sp,8],'-b',label=data_lbl[8][1])
                    if params.pnt.beta>0:
                        ax.plot(time,data[:,i_sp,9],'-m',label=data_lbl[9][1])
                        ax.plot(time,np.zeros_like(time),color='black',linewidth=1)
                    set_scales(ax,params)
                    ax.legend();
            figure.tight_layout()
            return figure
        for (a,b,c,d) in zip(self.time, self.data, self.params, self.dsets): 
            fig=plot_single(a,b,c,d) 
        
    
    def plot_fluctuations(self, window=None):
        if window:
            figure = window
        else:
            figure = plt.figure(figsize=(14,6), dpi= 100, facecolor='w', edgecolor='k')
        def plot_single(time, data, params, data_lbls):
            n_rows = 1
            n_cols = params.pnt.n_spec
            figure = plt.figure(figsize=set_figsize(), dpi= 100, facecolor='w', edgecolor='k')
            for i_sp in range(0, n_cols):  
                ax = figure.add_subplot(n_rows, n_cols, i_sp + 1)
                for i_dset in range(0,4):
                    ax.plot(time, data[:, i_sp, i_dset],label=data_lbls[i_dset][1])
                ax.set_xlabel(time_lbl(params)); ax.legend();ax.set_title(params.pnt.specnames[i_sp])
                set_scales(ax,params)
                ax.autoscale(enable=True, axis='y', tight=True)  
            figure.tight_layout()
            return figure
        for (a,b,c,d) in zip(self.time, self.data, self.params, self.dsets): 
            fig=plot_single(a,b,c,d) 
        
                     
    ## takes hyperparameters from first pnt, if hyperparameters vary through run, will need to pass extension of continuation
    def show_stats(self, ext=0, t_s=None, t_e=None, units=None):
        if not t_s:
            t_s=self.time[0]
        if not t_e:
            t_e=self.time[-1]
        t_s, i_s = find_nearest(self.time, t_s)     
        t_r, i_e = find_nearest(self.time, t_e)     
        for i_spec in range(self.pnt[ext].n_spec):
            print(self.pnt[ext].specnames[i_spec])
            for i_q, quant in enumerate(self.__set_dsets(self.pnt[ext].nrgcols)):
                res_tavg = mytrapz(self.data[i_s:i_e+1, i_spec, i_q], self.time[i_s:i_e+1])
                print("  {} = {:.2f}".format(quant, res_tavg))
                
        if units:
#            TODO
            for i_spec in range(self.pnt.n_spec):
                print(self.specnames[i_spec])
            for i_q, quant in enumerate(self.__set_dsets(self.pnt[ext].nrgcols)):
                res_tavg = mytrapz(self.data[i_s:i_e+1, i_spec, i_q], self.time[i_s:i_e+1])
                print("  {} = {:.2f}".format(quant, res_tavg)) 



# -*- coding: utf-8 -*-
from .run import Run
from ..data.data import Data
import os

class Simulation:
    ####################################################################
    ## The simulations class embeds the Run class,
    ##    meaning a simulation is a collection of runs
    ##    Aiming at something where a single run can be handled by itself
    ##   for cases to be used not in the GUI """
    ## in_folder [str/Path Obj]: directory gene output files reside
    ## out_folder [str]: directory to save results of diagnostics
    ## extensions [list(str/int)]: extensions of runs to include in diagnostics,
    ##                              pass as list iterable
    def __init__(self, in_folder=None, out_folder=None, extensions=None):
        # folder containing the simulation data
        self.in_folder = in_folder
        if in_folder:
            if not os.path.exists(in_folder):
                raise Exception("input folder does not exist")

        # extensions to be analyzed
        self.extensions = extensions

        # folder for output
        self.out_folder = out_folder
        if out_folder:
            if not os.path.exists(out_folder):
                raise Exception("output folder does not exist")

        # times to process
        self.starttime = -1
        self.endtime = -1

        # how often to step
        self.stepping = 1

        # how many steps to take
        self.max_steps = 'all'

        # this is the par_IO for each extension we have
        self.runs = []
        self.reset_runs()

    def reset_runs(self):
        """
        This method is meant for GUI usage mostly.
        If we already have a simulation object we will be changing folders
        (input, output) and runs to process). Each time we do that we need to
        redefine the object data
        """

        if not self.in_folder or not self.extensions:
            return

        if self.runs:
            self.runs.clear()

        for ext in self.extensions:
            self.runs.append(Run(self.in_folder, ext))


        """
         I put the data here in order to avoid memory duplication
         putting it inside runs, would make sense but continuation runs will
         keep track of data for nothing once they are processed.
         Still need to keep avail_vars and times per run.
         """
        
        self.data = Data(self.in_folder, self.extensions, self.runs)

    def __update_variables_and_times(self, variables, times):
        pass

